import subprocess
import time
from unittest import mock

import pytest

from briefcase.console import Console, Log
from briefcase.exceptions import BriefcaseCommandError
from briefcase.integrations.subprocess import Subprocess
from briefcase.integrations.xcode import DeviceState
from briefcase.platforms.iOS.xcode import iOSXcodeRunCommand
from briefcase.platforms.macOS import macOS_log_clean_filter


@pytest.fixture
def run_command(tmp_path):
    command = iOSXcodeRunCommand(
        logger=Log(),
        console=Console(),
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
    )
    command.tools.home_path = tmp_path / "home"
    command.tools.subprocess = mock.MagicMock(spec_set=Subprocess)
    command._stream_app_logs = mock.MagicMock()

    # To satisfy coverage, the stop function must be invoked
    # at least once when streaming app logs.
    def mock_stream_app_logs(app, stop_func, **kwargs):
        stop_func()

    command._stream_app_logs.side_effect = mock_stream_app_logs

    return command


def test_device_option(run_command):
    """The -d option can be parsed."""
    options, overrides = run_command.parse_options(["-d", "myphone"])

    assert options == {
        "udid": "myphone",
        "update": False,
        "update_requirements": False,
        "update_resources": False,
        "update_support": False,
        "update_stub": False,
        "no_update": False,
        "test_mode": False,
        "passthrough": [],
        "appname": None,
    }
    assert overrides == {}


def test_run_multiple_devices_input_disabled(run_command, first_app_config):
    """If input is disabled, but there are multiple devices, an error is raised."""
    # Multiple devices are available
    run_command.get_simulators = mock.MagicMock(
        return_value={
            "iOS 13.2": {
                "C9A005C8-9468-47C5-8376-68A6E3408209": "iPhone 8",
                "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D": "iPhone 11",
                "EEEBA06C-81F9-407C-885A-2261306DB2BE": "iPhone 11 Pro Max",
            }
        }
    )

    # Disable console input.
    run_command.tools.input.enabled = False

    with pytest.raises(
        BriefcaseCommandError,
        match=r"Input has been disabled; can't select a device to target.",
    ):
        run_command.run_app(first_app_config, test_mode=False, passthrough=[])


@pytest.mark.usefixtures("sleep_zero")
def test_run_app_simulator_booted(run_command, first_app_config, tmp_path):
    """An iOS App can be started when the simulator is already booted."""
    # A valid target device will be selected.
    run_command.select_target_device = mock.MagicMock(
        return_value=("2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D", "13.2", "iPhone 11")
    )

    # Simulator is already booted
    run_command.get_device_state = mock.MagicMock(return_value=DeviceState.BOOTED)

    # Mock a process ID for the app
    run_command.tools.subprocess.check_output.return_value = (
        "com.example.first-app: 1234\n"
    )

    # Mock the uninstall, install, and log stream Popen processes
    uninstall_popen = mock.MagicMock(spec_set=subprocess.Popen)
    uninstall_popen.__enter__.return_value = uninstall_popen
    uninstall_popen.poll.side_effect = [None, None, 0]
    install_popen = mock.MagicMock(spec_set=subprocess.Popen)
    install_popen.__enter__.return_value = install_popen
    install_popen.poll.side_effect = [None, None, 0]
    log_stream_process = mock.MagicMock(spec_set=subprocess.Popen)
    run_command.tools.subprocess.Popen.side_effect = [
        uninstall_popen,
        install_popen,
        log_stream_process,
    ]

    # Run the app
    run_command.run_app(first_app_config, test_mode=False, passthrough=[])

    # The correct sequence of commands was issued.
    run_command.tools.subprocess.run.assert_has_calls(
        [
            # Open the simulator
            mock.call(
                [
                    "open",
                    "-a",
                    "Simulator",
                    "--args",
                    "-CurrentDeviceUDID",
                    "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D",
                ],
                check=True,
            ),
        ]
    )
    # Launch the new app
    run_command.tools.subprocess.check_output.assert_called_once_with(
        [
            "xcrun",
            "simctl",
            "launch",
            "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D",
            "com.example.first-app",
        ],
    )

    # slept 4 times for uninstall/install and 1 time for log stream start
    assert time.sleep.call_count == 4 + 1

    # Start the uninstall, install, and  log stream
    run_command.tools.subprocess.Popen.assert_has_calls(
        [
            # Uninstall the old app
            mock.call(
                [
                    "xcrun",
                    "simctl",
                    "uninstall",
                    "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D",
                    "com.example.first-app",
                ],
            ),
            # Install the new app
            mock.call(
                [
                    "xcrun",
                    "simctl",
                    "install",
                    "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D",
                    tmp_path
                    / "base_path/build/first-app/ios/xcode/build/Debug-iphonesimulator/First App.app",
                ],
            ),
            mock.call(
                [
                    "xcrun",
                    "simctl",
                    "spawn",
                    "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D",
                    "log",
                    "stream",
                    "--style",
                    "compact",
                    "--predicate",
                    'senderImagePath ENDSWITH "/First App"'
                    ' OR (processImagePath ENDSWITH "/First App"'
                    ' AND (senderImagePath ENDSWITH "-iphonesimulator.so"'
                    ' OR senderImagePath ENDSWITH "-iphonesimulator.dylib"))',
                ],
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                bufsize=1,
            ),
        ]
    )

    # Log stream monitoring was started
    run_command._stream_app_logs.assert_called_with(
        first_app_config,
        popen=log_stream_process,
        test_mode=False,
        clean_filter=macOS_log_clean_filter,
        clean_output=True,
        stop_func=mock.ANY,
        log_stream=True,
    )


@pytest.mark.usefixtures("sleep_zero")
def test_run_app_simulator_booted_underscore(
    run_command,
    underscore_app_config,
    tmp_path,
):
    """An iOS App can be started when the simulator is already booted.

    This test is specific to app names that have underscores since those are not
    supported by Apple within app identifiers.
    """
    # A valid target device will be selected.
    run_command.select_target_device = mock.MagicMock(
        return_value=("2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D", "13.2", "iPhone 11")
    )

    # Simulator is already booted
    run_command.get_device_state = mock.MagicMock(return_value=DeviceState.BOOTED)

    # Mock a process ID for the app
    run_command.tools.subprocess.check_output.return_value = (
        "com.example.first-app: 1234\n"
    )

    # Mock the uninstall, install, and log stream Popen processes
    uninstall_popen = mock.MagicMock(spec_set=subprocess.Popen)
    uninstall_popen.__enter__.return_value = uninstall_popen
    uninstall_popen.poll.side_effect = [None, None, 0]
    install_popen = mock.MagicMock(spec_set=subprocess.Popen)
    install_popen.__enter__.return_value = install_popen
    install_popen.poll.side_effect = [None, None, 0]
    log_stream_process = mock.MagicMock(spec_set=subprocess.Popen)
    run_command.tools.subprocess.Popen.side_effect = [
        uninstall_popen,
        install_popen,
        log_stream_process,
    ]

    # Run the app
    run_command.run_app(underscore_app_config, test_mode=False, passthrough=[])

    # slept 4 times for uninstall/install and 1 time for log stream start
    assert time.sleep.call_count == 4 + 1

    # The correct sequence of commands was issued.
    run_command.tools.subprocess.run.assert_has_calls(
        [
            # Open the simulator
            mock.call(
                [
                    "open",
                    "-a",
                    "Simulator",
                    "--args",
                    "-CurrentDeviceUDID",
                    "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D",
                ],
                check=True,
            ),
        ]
    )
    # Launch the new app
    run_command.tools.subprocess.check_output.assert_called_once_with(
        [
            "xcrun",
            "simctl",
            "launch",
            "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D",
            "com.example.first-app",
        ],
    )

    # Start the uninstall, install, and  log stream
    run_command.tools.subprocess.Popen.assert_has_calls(
        [
            # Uninstall the old app
            mock.call(
                [
                    "xcrun",
                    "simctl",
                    "uninstall",
                    "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D",
                    "com.example.first-app",
                ],
            ),
            # Install the new app
            mock.call(
                [
                    "xcrun",
                    "simctl",
                    "install",
                    "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D",
                    tmp_path
                    / "base_path/build/first_app/ios/xcode/build/Debug-iphonesimulator/First App.app",
                ],
            ),
            mock.call(
                [
                    "xcrun",
                    "simctl",
                    "spawn",
                    "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D",
                    "log",
                    "stream",
                    "--style",
                    "compact",
                    "--predicate",
                    'senderImagePath ENDSWITH "/First App"'
                    ' OR (processImagePath ENDSWITH "/First App"'
                    ' AND (senderImagePath ENDSWITH "-iphonesimulator.so"'
                    ' OR senderImagePath ENDSWITH "-iphonesimulator.dylib"))',
                ],
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                bufsize=1,
            ),
        ]
    )

    # Log stream monitoring was started
    run_command._stream_app_logs.assert_called_with(
        underscore_app_config,
        popen=log_stream_process,
        test_mode=False,
        clean_filter=macOS_log_clean_filter,
        clean_output=True,
        stop_func=mock.ANY,
        log_stream=True,
    )


@pytest.mark.usefixtures("sleep_zero")
def test_run_app_with_passthrough(run_command, first_app_config, tmp_path):
    """An iOS App can be started with passthrough args."""
    # A valid target device will be selected.
    run_command.select_target_device = mock.MagicMock(
        return_value=("2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D", "13.2", "iPhone 11")
    )

    # Simulator is already booted
    run_command.get_device_state = mock.MagicMock(return_value=DeviceState.BOOTED)

    # Mock a process ID for the app
    run_command.tools.subprocess.check_output.return_value = (
        "com.example.first-app: 1234\n"
    )

    # Mock the uninstall, install, and log stream Popen processes
    uninstall_popen = mock.MagicMock(spec_set=subprocess.Popen)
    uninstall_popen.__enter__.return_value = uninstall_popen
    uninstall_popen.poll.side_effect = [None, None, 0]
    install_popen = mock.MagicMock(spec_set=subprocess.Popen)
    install_popen.__enter__.return_value = install_popen
    install_popen.poll.side_effect = [None, None, 0]
    log_stream_process = mock.MagicMock(spec_set=subprocess.Popen)
    run_command.tools.subprocess.Popen.side_effect = [
        uninstall_popen,
        install_popen,
        log_stream_process,
    ]

    # Run the app with passthrough args.
    run_command.run_app(
        first_app_config,
        test_mode=False,
        passthrough=["foo", "--bar"],
    )

    # slept 4 times for uninstall/install and 1 time for log stream start
    assert time.sleep.call_count == 4 + 1

    # The correct sequence of commands was issued.
    run_command.tools.subprocess.run.assert_has_calls(
        [
            # Open the simulator
            mock.call(
                [
                    "open",
                    "-a",
                    "Simulator",
                    "--args",
                    "-CurrentDeviceUDID",
                    "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D",
                ],
                check=True,
            ),
        ]
    )
    # Launch the new app
    run_command.tools.subprocess.check_output.assert_called_once_with(
        [
            "xcrun",
            "simctl",
            "launch",
            "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D",
            "com.example.first-app",
            "foo",
            "--bar",
        ],
    )

    # Start the uninstall, install, and  log stream
    run_command.tools.subprocess.Popen.assert_has_calls(
        [
            # Uninstall the old app
            mock.call(
                [
                    "xcrun",
                    "simctl",
                    "uninstall",
                    "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D",
                    "com.example.first-app",
                ],
            ),
            # Install the new app
            mock.call(
                [
                    "xcrun",
                    "simctl",
                    "install",
                    "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D",
                    tmp_path
                    / "base_path/build/first-app/ios/xcode/build/Debug-iphonesimulator/First App.app",
                ],
            ),
            mock.call(
                [
                    "xcrun",
                    "simctl",
                    "spawn",
                    "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D",
                    "log",
                    "stream",
                    "--style",
                    "compact",
                    "--predicate",
                    'senderImagePath ENDSWITH "/First App"'
                    ' OR (processImagePath ENDSWITH "/First App"'
                    ' AND (senderImagePath ENDSWITH "-iphonesimulator.so"'
                    ' OR senderImagePath ENDSWITH "-iphonesimulator.dylib"))',
                ],
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                bufsize=1,
            ),
        ]
    )

    # Log stream monitoring was started
    run_command._stream_app_logs.assert_called_with(
        first_app_config,
        popen=log_stream_process,
        test_mode=False,
        clean_filter=macOS_log_clean_filter,
        clean_output=True,
        stop_func=mock.ANY,
        log_stream=True,
    )


@pytest.mark.usefixtures("sleep_zero")
def test_run_app_simulator_shut_down(
    run_command,
    first_app_config,
    tmp_path,
):
    """An iOS App can be started when the simulator is shut down."""
    # A valid target device will be selected.
    run_command.select_target_device = mock.MagicMock(
        return_value=("2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D", "13.2", "iPhone 11")
    )

    # Simulator is shut down
    run_command.get_device_state = mock.MagicMock(return_value=DeviceState.SHUTDOWN)

    run_command.tools.subprocess = mock.MagicMock(spec_set=Subprocess)

    # Mock a process ID for the app
    run_command.tools.subprocess.check_output.return_value = (
        "com.example.first-app: 1234\n"
    )

    # Mock the uninstall, install, and log stream Popen processes
    uninstall_popen = mock.MagicMock(spec_set=subprocess.Popen)
    uninstall_popen.__enter__.return_value = uninstall_popen
    uninstall_popen.poll.side_effect = [None, None, 0]
    install_popen = mock.MagicMock(spec_set=subprocess.Popen)
    install_popen.__enter__.return_value = install_popen
    install_popen.poll.side_effect = [None, None, 0]
    log_stream_process = mock.MagicMock(spec_set=subprocess.Popen)
    run_command.tools.subprocess.Popen.side_effect = [
        uninstall_popen,
        install_popen,
        log_stream_process,
    ]

    # Run the app
    run_command.run_app(first_app_config, test_mode=False, passthrough=[])

    # slept 4 times for uninstall/install and 1 time for log stream start
    assert time.sleep.call_count == 4 + 1

    # The correct sequence of commands was issued.
    run_command.tools.subprocess.run.assert_has_calls(
        [
            # Boot the device
            mock.call(
                ["xcrun", "simctl", "boot", "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D"],
                check=True,
            ),
            # Open the simulator
            mock.call(
                [
                    "open",
                    "-a",
                    "Simulator",
                    "--args",
                    "-CurrentDeviceUDID",
                    "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D",
                ],
                check=True,
            ),
        ]
    )

    # Launch the new app
    run_command.tools.subprocess.check_output.assert_called_once_with(
        [
            "xcrun",
            "simctl",
            "launch",
            "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D",
            "com.example.first-app",
        ],
    )

    # Start the uninstall, install, and  log stream
    run_command.tools.subprocess.Popen.assert_has_calls(
        [
            # Uninstall the old app
            mock.call(
                [
                    "xcrun",
                    "simctl",
                    "uninstall",
                    "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D",
                    "com.example.first-app",
                ],
            ),
            # Install the new app
            mock.call(
                [
                    "xcrun",
                    "simctl",
                    "install",
                    "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D",
                    tmp_path
                    / "base_path/build/first-app/ios/xcode/build/Debug-iphonesimulator/First App.app",
                ],
            ),
            mock.call(
                [
                    "xcrun",
                    "simctl",
                    "spawn",
                    "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D",
                    "log",
                    "stream",
                    "--style",
                    "compact",
                    "--predicate",
                    'senderImagePath ENDSWITH "/First App"'
                    ' OR (processImagePath ENDSWITH "/First App"'
                    ' AND (senderImagePath ENDSWITH "-iphonesimulator.so"'
                    ' OR senderImagePath ENDSWITH "-iphonesimulator.dylib"))',
                ],
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                bufsize=1,
            ),
        ]
    )

    # Log stream monitoring was started
    run_command._stream_app_logs.assert_called_with(
        first_app_config,
        popen=log_stream_process,
        test_mode=False,
        clean_filter=macOS_log_clean_filter,
        clean_output=True,
        stop_func=mock.ANY,
        log_stream=True,
    )


@pytest.mark.usefixtures("sleep_zero")
def test_run_app_simulator_shutting_down(run_command, first_app_config, tmp_path):
    """An iOS App can be started when the simulator is shutting down."""
    # A valid target device will be selected.
    run_command.select_target_device = mock.MagicMock(
        return_value=("2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D", "13.2", "iPhone 11")
    )

    # Simulator is shutting down. This will be returned a couple of times,
    # as the simulator will take a few seconds before it is fully shut down.
    # There will be a sleep between each call, so we need to mock sleep as well.
    run_command.get_device_state = mock.MagicMock(
        side_effect=[
            DeviceState.SHUTTING_DOWN,
            DeviceState.SHUTTING_DOWN,
            DeviceState.SHUTTING_DOWN,
            DeviceState.SHUTDOWN,
        ]
    )

    run_command.sleep = mock.MagicMock()
    run_command.tools.subprocess = mock.MagicMock(spec_set=Subprocess)

    # Mock a process ID for the app
    run_command.tools.subprocess.check_output.return_value = (
        "com.example.first-app: 1234\n"
    )

    # Mock the uninstall, install, and log stream Popen processes
    uninstall_popen = mock.MagicMock(spec_set=subprocess.Popen)
    uninstall_popen.__enter__.return_value = uninstall_popen
    uninstall_popen.poll.side_effect = [None, None, 0]
    install_popen = mock.MagicMock(spec_set=subprocess.Popen)
    install_popen.__enter__.return_value = install_popen
    install_popen.poll.side_effect = [None, None, 0]
    log_stream_process = mock.MagicMock(spec_set=subprocess.Popen)
    run_command.tools.subprocess.Popen.side_effect = [
        uninstall_popen,
        install_popen,
        log_stream_process,
    ]

    # Run the app
    run_command.run_app(first_app_config, test_mode=False, passthrough=[])

    # We should have slept 4 times for shutting down and 4 time for uninstall/install
    assert time.sleep.call_count == 4 + 4

    # The correct sequence of commands was issued.
    run_command.tools.subprocess.run.assert_has_calls(
        [
            # Boot the device
            mock.call(
                ["xcrun", "simctl", "boot", "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D"],
                check=True,
            ),
            # Open the simulator
            mock.call(
                [
                    "open",
                    "-a",
                    "Simulator",
                    "--args",
                    "-CurrentDeviceUDID",
                    "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D",
                ],
                check=True,
            ),
        ]
    )

    # Launch the new app
    run_command.tools.subprocess.check_output.assert_called_once_with(
        [
            "xcrun",
            "simctl",
            "launch",
            "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D",
            "com.example.first-app",
        ],
    )

    # Start the uninstall, install, and  log stream
    run_command.tools.subprocess.Popen.assert_has_calls(
        [
            # Uninstall the old app
            mock.call(
                [
                    "xcrun",
                    "simctl",
                    "uninstall",
                    "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D",
                    "com.example.first-app",
                ],
            ),
            # Install the new app
            mock.call(
                [
                    "xcrun",
                    "simctl",
                    "install",
                    "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D",
                    tmp_path
                    / "base_path/build/first-app/ios/xcode/build/Debug-iphonesimulator/First App.app",
                ],
            ),
            mock.call(
                [
                    "xcrun",
                    "simctl",
                    "spawn",
                    "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D",
                    "log",
                    "stream",
                    "--style",
                    "compact",
                    "--predicate",
                    'senderImagePath ENDSWITH "/First App"'
                    ' OR (processImagePath ENDSWITH "/First App"'
                    ' AND (senderImagePath ENDSWITH "-iphonesimulator.so"'
                    ' OR senderImagePath ENDSWITH "-iphonesimulator.dylib"))',
                ],
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                bufsize=1,
            ),
        ]
    )

    # Log stream monitoring was started
    run_command._stream_app_logs.assert_called_with(
        first_app_config,
        popen=log_stream_process,
        test_mode=False,
        clean_filter=macOS_log_clean_filter,
        clean_output=True,
        stop_func=mock.ANY,
        log_stream=True,
    )


@pytest.mark.usefixtures("sleep_zero")
def test_run_app_simulator_boot_failure(run_command, first_app_config):
    """If the simulator fails to boot, raise an error."""
    # A valid target device will be selected.
    run_command.select_target_device = mock.MagicMock(
        return_value=("2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D", "13.2", "iPhone 11")
    )

    # Simulator is shut down
    run_command.get_device_state = mock.MagicMock(return_value=DeviceState.SHUTDOWN)

    run_command.tools.subprocess = mock.MagicMock(spec_set=Subprocess)
    run_command.tools.subprocess.run.side_effect = subprocess.CalledProcessError(
        cmd=["xcrun", "simclt", "boot", "..."], returncode=1
    )

    # Run the app
    with pytest.raises(BriefcaseCommandError):
        run_command.run_app(first_app_config, test_mode=False, passthrough=[])

    # No sleeps
    assert time.sleep.call_count == 0

    # The correct sequence of commands was issued.
    run_command.tools.subprocess.run.assert_has_calls(
        [
            # Boot the device
            mock.call(
                ["xcrun", "simctl", "boot", "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D"],
                check=True,
            ),
        ]
    )

    # The log will not be tailed
    run_command.tools.subprocess.Popen.assert_not_called()
    run_command._stream_app_logs.assert_not_called()


@pytest.mark.usefixtures("sleep_zero")
def test_run_app_simulator_open_failure(run_command, first_app_config):
    """If the simulator can't be opened, raise an error."""
    # A valid target device will be selected.
    run_command.select_target_device = mock.MagicMock(
        return_value=("2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D", "13.2", "iPhone 11")
    )

    # Simulator is shut down
    run_command.get_device_state = mock.MagicMock(return_value=DeviceState.SHUTDOWN)

    # Call to boot succeeds, but open fails.
    run_command.tools.subprocess = mock.MagicMock(spec_set=Subprocess)
    run_command.tools.subprocess.run.side_effect = [
        0,
        subprocess.CalledProcessError(
            cmd=["open", "-a", "Simulator", "..."],
            returncode=1,
        ),
    ]

    # Run the app
    with pytest.raises(BriefcaseCommandError):
        run_command.run_app(first_app_config, test_mode=False, passthrough=[])

    # No sleeps
    assert time.sleep.call_count == 0

    # The correct sequence of commands was issued.
    run_command.tools.subprocess.run.assert_has_calls(
        [
            # Boot the device
            mock.call(
                ["xcrun", "simctl", "boot", "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D"],
                check=True,
            ),
            # Open the simulator
            mock.call(
                [
                    "open",
                    "-a",
                    "Simulator",
                    "--args",
                    "-CurrentDeviceUDID",
                    "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D",
                ],
                check=True,
            ),
        ]
    )
    # The log will not be tailed
    run_command.tools.subprocess.Popen.assert_not_called()
    run_command._stream_app_logs.assert_not_called()


@pytest.mark.usefixtures("sleep_zero")
def test_run_app_simulator_uninstall_failure(run_command, first_app_config):
    """If the old app can't be uninstalled, raise an error."""
    # A valid target device will be selected.
    run_command.select_target_device = mock.MagicMock(
        return_value=("2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D", "13.2", "iPhone 11")
    )

    # Simulator is shut down
    run_command.get_device_state = mock.MagicMock(return_value=DeviceState.SHUTDOWN)

    # Call uninstall fails
    uninstall_popen = mock.MagicMock(spec_set=subprocess.Popen)
    uninstall_popen.__enter__.return_value = uninstall_popen
    uninstall_popen.poll.side_effect = [None, None, 1]
    run_command.tools.subprocess.Popen.side_effect = [uninstall_popen]

    # Run the app
    with pytest.raises(BriefcaseCommandError):
        run_command.run_app(first_app_config, test_mode=False, passthrough=[])

    # Sleep twice for uninstall failure
    assert time.sleep.call_count == 2

    # The correct sequence of commands was issued.
    run_command.tools.subprocess.run.assert_has_calls(
        [
            # Boot the device
            mock.call(
                ["xcrun", "simctl", "boot", "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D"],
                check=True,
            ),
            # Open the simulator
            mock.call(
                [
                    "open",
                    "-a",
                    "Simulator",
                    "--args",
                    "-CurrentDeviceUDID",
                    "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D",
                ],
                check=True,
            ),
        ]
    )
    # Start the uninstall
    run_command.tools.subprocess.Popen.assert_has_calls(
        [
            # Uninstall the old app
            mock.call(
                [
                    "xcrun",
                    "simctl",
                    "uninstall",
                    "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D",
                    "com.example.first-app",
                ],
            ),
        ]
    )

    # The log will not be tailed
    run_command._stream_app_logs.assert_not_called()


@pytest.mark.usefixtures("sleep_zero")
def test_run_app_simulator_install_failure(run_command, first_app_config, tmp_path):
    """If the app fails to install in the simulator, raise an error."""
    # A valid target device will be selected.
    run_command.select_target_device = mock.MagicMock(
        return_value=("2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D", "13.2", "iPhone 11")
    )

    # Simulator is shut down
    run_command.get_device_state = mock.MagicMock(return_value=DeviceState.SHUTDOWN)

    # Call to install fails
    uninstall_popen = mock.MagicMock(spec_set=subprocess.Popen)
    uninstall_popen.__enter__.return_value = uninstall_popen
    uninstall_popen.poll.side_effect = [None, None, 0]
    install_popen = mock.MagicMock(spec_set=subprocess.Popen)
    install_popen.__enter__.return_value = install_popen
    install_popen.poll.side_effect = [None, None, 1]
    run_command.tools.subprocess.Popen.side_effect = [
        uninstall_popen,
        install_popen,
    ]

    # Run the app
    with pytest.raises(BriefcaseCommandError):
        run_command.run_app(first_app_config, test_mode=False, passthrough=[])

    # Sleep twice for uninstall and twice for install failure
    assert time.sleep.call_count == 4

    # The correct sequence of commands was issued.
    run_command.tools.subprocess.run.assert_has_calls(
        [
            # Boot the device
            mock.call(
                ["xcrun", "simctl", "boot", "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D"],
                check=True,
            ),
            # Open the simulator
            mock.call(
                [
                    "open",
                    "-a",
                    "Simulator",
                    "--args",
                    "-CurrentDeviceUDID",
                    "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D",
                ],
                check=True,
            ),
        ]
    )

    # Start the uninstall and install
    run_command.tools.subprocess.Popen.assert_has_calls(
        [
            # Uninstall the old app
            mock.call(
                [
                    "xcrun",
                    "simctl",
                    "uninstall",
                    "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D",
                    "com.example.first-app",
                ],
            ),
            # Install the new app
            mock.call(
                [
                    "xcrun",
                    "simctl",
                    "install",
                    "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D",
                    tmp_path
                    / "base_path/build/first-app/ios/xcode/build/Debug-iphonesimulator/First App.app",
                ],
            ),
        ]
    )

    # The log will not be tailed
    run_command._stream_app_logs.assert_not_called()


@pytest.mark.usefixtures("sleep_zero")
def test_run_app_simulator_launch_failure(run_command, first_app_config, tmp_path):
    """If the app fails to launch, raise an error."""
    # A valid target device will be selected.
    run_command.select_target_device = mock.MagicMock(
        return_value=("2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D", "13.2", "iPhone 11")
    )

    # Simulator is shut down
    run_command.get_device_state = mock.MagicMock(return_value=DeviceState.SHUTDOWN)

    # Mock a process ID for the app
    run_command.tools.subprocess.check_output.side_effect = (
        subprocess.CalledProcessError(
            cmd=["xcrun", "simctl", "launch", "..."], returncode=1
        )
    )

    # Mock the uninstall, install, and log stream Popen processes
    uninstall_popen = mock.MagicMock(spec_set=subprocess.Popen)
    uninstall_popen.__enter__.return_value = uninstall_popen
    uninstall_popen.poll.side_effect = [None, None, 0]
    install_popen = mock.MagicMock(spec_set=subprocess.Popen)
    install_popen.__enter__.return_value = install_popen
    install_popen.poll.side_effect = [None, None, 0]
    log_stream_process = mock.MagicMock(spec_set=subprocess.Popen)
    run_command.tools.subprocess.Popen.side_effect = [
        uninstall_popen,
        install_popen,
        log_stream_process,
    ]

    # Run the app
    with pytest.raises(BriefcaseCommandError):
        run_command.run_app(first_app_config, test_mode=False, passthrough=[])

    # Sleep four times for uninstall/install and once for log stream start
    assert time.sleep.call_count == 4 + 1

    # The correct sequence of commands was issued.
    run_command.tools.subprocess.run.assert_has_calls(
        [
            # Boot the device
            mock.call(
                ["xcrun", "simctl", "boot", "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D"],
                check=True,
            ),
            # Open the simulator
            mock.call(
                [
                    "open",
                    "-a",
                    "Simulator",
                    "--args",
                    "-CurrentDeviceUDID",
                    "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D",
                ],
                check=True,
            ),
        ]
    )

    # Launch the new app
    run_command.tools.subprocess.check_output.assert_called_once_with(
        [
            "xcrun",
            "simctl",
            "launch",
            "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D",
            "com.example.first-app",
        ],
    )

    # Start the uninstall, install, and  log stream
    run_command.tools.subprocess.Popen.assert_has_calls(
        [
            # Uninstall the old app
            mock.call(
                [
                    "xcrun",
                    "simctl",
                    "uninstall",
                    "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D",
                    "com.example.first-app",
                ],
            ),
            # Install the new app
            mock.call(
                [
                    "xcrun",
                    "simctl",
                    "install",
                    "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D",
                    tmp_path
                    / "base_path/build/first-app/ios/xcode/build/Debug-iphonesimulator/First App.app",
                ],
            ),
            mock.call(
                [
                    "xcrun",
                    "simctl",
                    "spawn",
                    "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D",
                    "log",
                    "stream",
                    "--style",
                    "compact",
                    "--predicate",
                    'senderImagePath ENDSWITH "/First App"'
                    ' OR (processImagePath ENDSWITH "/First App"'
                    ' AND (senderImagePath ENDSWITH "-iphonesimulator.so"'
                    ' OR senderImagePath ENDSWITH "-iphonesimulator.dylib"))',
                ],
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                bufsize=1,
            ),
        ]
    )

    # Log stream failed, so it won't be monitored
    run_command._stream_app_logs.assert_not_called()


@pytest.mark.usefixtures("sleep_zero")
def test_run_app_simulator_no_pid(run_command, first_app_config, tmp_path):
    """If the app fails to provide a meaningful PID on launch, raise an error."""
    # A valid target device will be selected.
    run_command.select_target_device = mock.MagicMock(
        return_value=("2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D", "13.2", "iPhone 11")
    )

    # Simulator is shut down
    run_command.get_device_state = mock.MagicMock(return_value=DeviceState.SHUTDOWN)

    # Mock a bad return value for the PID
    run_command.tools.subprocess.check_output.return_value = "No PID returned"

    # Mock the uninstall, install, and log stream Popen processes
    uninstall_popen = mock.MagicMock(spec_set=subprocess.Popen)
    uninstall_popen.__enter__.return_value = uninstall_popen
    uninstall_popen.poll.side_effect = [None, None, 0]
    install_popen = mock.MagicMock(spec_set=subprocess.Popen)
    install_popen.__enter__.return_value = install_popen
    install_popen.poll.side_effect = [None, None, 0]
    log_stream_process = mock.MagicMock(spec_set=subprocess.Popen)
    run_command.tools.subprocess.Popen.side_effect = [
        uninstall_popen,
        install_popen,
        log_stream_process,
    ]

    # Run the app
    with pytest.raises(BriefcaseCommandError):
        run_command.run_app(first_app_config, test_mode=False, passthrough=[])

    # Sleep four times for uninstall/install and once for log stream start
    assert time.sleep.call_count == 4 + 1

    # The correct sequence of commands was issued.
    run_command.tools.subprocess.run.assert_has_calls(
        [
            # Boot the device
            mock.call(
                ["xcrun", "simctl", "boot", "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D"],
                check=True,
            ),
            # Open the simulator
            mock.call(
                [
                    "open",
                    "-a",
                    "Simulator",
                    "--args",
                    "-CurrentDeviceUDID",
                    "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D",
                ],
                check=True,
            ),
        ]
    )

    # Launch the new app
    run_command.tools.subprocess.check_output.assert_called_once_with(
        [
            "xcrun",
            "simctl",
            "launch",
            "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D",
            "com.example.first-app",
        ],
    )

    # Start the uninstall, install, and  log stream
    run_command.tools.subprocess.Popen.assert_has_calls(
        [
            # Uninstall the old app
            mock.call(
                [
                    "xcrun",
                    "simctl",
                    "uninstall",
                    "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D",
                    "com.example.first-app",
                ],
            ),
            # Install the new app
            mock.call(
                [
                    "xcrun",
                    "simctl",
                    "install",
                    "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D",
                    tmp_path
                    / "base_path/build/first-app/ios/xcode/build/Debug-iphonesimulator/First App.app",
                ],
            ),
            mock.call(
                [
                    "xcrun",
                    "simctl",
                    "spawn",
                    "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D",
                    "log",
                    "stream",
                    "--style",
                    "compact",
                    "--predicate",
                    'senderImagePath ENDSWITH "/First App"'
                    ' OR (processImagePath ENDSWITH "/First App"'
                    ' AND (senderImagePath ENDSWITH "-iphonesimulator.so"'
                    ' OR senderImagePath ENDSWITH "-iphonesimulator.dylib"))',
                ],
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                bufsize=1,
            ),
        ]
    )

    # PID detection failed, so it won't be monitored
    run_command._stream_app_logs.assert_not_called()


@pytest.mark.usefixtures("sleep_zero")
def test_run_app_simulator_non_integer_pid(run_command, first_app_config, tmp_path):
    """If the PID returned isn't an integer, raise an error."""
    # A valid target device will be selected.
    run_command.select_target_device = mock.MagicMock(
        return_value=("2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D", "13.2", "iPhone 11")
    )

    # Simulator is shut down
    run_command.get_device_state = mock.MagicMock(return_value=DeviceState.SHUTDOWN)

    # Mock a bad process ID for the app
    run_command.tools.subprocess.check_output.return_value = (
        "com.example.first-app: NOT A PID\n"
    )

    # Mock the uninstall, install, and log stream Popen processes
    uninstall_popen = mock.MagicMock(spec_set=subprocess.Popen)
    uninstall_popen.__enter__.return_value = uninstall_popen
    uninstall_popen.poll.side_effect = [None, None, 0]
    install_popen = mock.MagicMock(spec_set=subprocess.Popen)
    install_popen.__enter__.return_value = install_popen
    install_popen.poll.side_effect = [None, None, 0]
    log_stream_process = mock.MagicMock(spec_set=subprocess.Popen)
    run_command.tools.subprocess.Popen.side_effect = [
        uninstall_popen,
        install_popen,
        log_stream_process,
    ]

    # Run the app
    with pytest.raises(BriefcaseCommandError):
        run_command.run_app(first_app_config, test_mode=False, passthrough=[])

    # Sleep four times for uninstall/install and once for log stream start
    assert time.sleep.call_count == 4 + 1

    # The correct sequence of commands was issued.
    run_command.tools.subprocess.run.assert_has_calls(
        [
            # Boot the device
            mock.call(
                ["xcrun", "simctl", "boot", "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D"],
                check=True,
            ),
            # Open the simulator
            mock.call(
                [
                    "open",
                    "-a",
                    "Simulator",
                    "--args",
                    "-CurrentDeviceUDID",
                    "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D",
                ],
                check=True,
            ),
        ]
    )

    # Launch the new app
    run_command.tools.subprocess.check_output.assert_called_once_with(
        [
            "xcrun",
            "simctl",
            "launch",
            "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D",
            "com.example.first-app",
        ],
    )

    # Start the uninstall, install, and  log stream
    run_command.tools.subprocess.Popen.assert_has_calls(
        [
            # Uninstall the old app
            mock.call(
                [
                    "xcrun",
                    "simctl",
                    "uninstall",
                    "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D",
                    "com.example.first-app",
                ],
            ),
            # Install the new app
            mock.call(
                [
                    "xcrun",
                    "simctl",
                    "install",
                    "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D",
                    tmp_path
                    / "base_path/build/first-app/ios/xcode/build/Debug-iphonesimulator/First App.app",
                ],
            ),
            mock.call(
                [
                    "xcrun",
                    "simctl",
                    "spawn",
                    "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D",
                    "log",
                    "stream",
                    "--style",
                    "compact",
                    "--predicate",
                    'senderImagePath ENDSWITH "/First App"'
                    ' OR (processImagePath ENDSWITH "/First App"'
                    ' AND (senderImagePath ENDSWITH "-iphonesimulator.so"'
                    ' OR senderImagePath ENDSWITH "-iphonesimulator.dylib"))',
                ],
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                bufsize=1,
            ),
        ]
    )

    # PID detection failed, so it won't be called
    run_command._stream_app_logs.assert_not_called()


@pytest.mark.usefixtures("sleep_zero")
def test_run_app_test_mode(run_command, first_app_config, tmp_path):
    """An iOS App can be started in test mode."""
    # A valid target device will be selected.
    run_command.select_target_device = mock.MagicMock(
        return_value=("2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D", "13.2", "iPhone 11")
    )

    # Simulator is already booted
    run_command.get_device_state = mock.MagicMock(return_value=DeviceState.BOOTED)

    # Mock a process ID for the app
    run_command.tools.subprocess.check_output.return_value = (
        "com.example.first-app: 1234\n"
    )

    # Mock the uninstall, install, and log stream Popen processes
    uninstall_popen = mock.MagicMock(spec_set=subprocess.Popen)
    uninstall_popen.__enter__.return_value = uninstall_popen
    uninstall_popen.poll.side_effect = [None, None, 0]
    install_popen = mock.MagicMock(spec_set=subprocess.Popen)
    install_popen.__enter__.return_value = install_popen
    install_popen.poll.side_effect = [None, None, 0]
    log_stream_process = mock.MagicMock(spec_set=subprocess.Popen)
    run_command.tools.subprocess.Popen.side_effect = [
        uninstall_popen,
        install_popen,
        log_stream_process,
    ]

    # Run the app
    run_command.run_app(first_app_config, test_mode=True, passthrough=[])

    # Sleep four times for uninstall/install and once for log stream start
    assert time.sleep.call_count == 4 + 1

    # Open command was not called
    run_command.tools.subprocess.run.assert_not_called()

    # Launch the new app
    run_command.tools.subprocess.check_output.assert_called_once_with(
        [
            "xcrun",
            "simctl",
            "launch",
            "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D",
            "com.example.first-app",
        ],
    )

    # Start the uninstall, install, and  log stream
    run_command.tools.subprocess.Popen.assert_has_calls(
        [
            # Uninstall the old app
            mock.call(
                [
                    "xcrun",
                    "simctl",
                    "uninstall",
                    "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D",
                    "com.example.first-app",
                ],
            ),
            # Install the new app
            mock.call(
                [
                    "xcrun",
                    "simctl",
                    "install",
                    "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D",
                    tmp_path
                    / "base_path/build/first-app/ios/xcode/build/Debug-iphonesimulator/First App.app",
                ],
            ),
            mock.call(
                [
                    "xcrun",
                    "simctl",
                    "spawn",
                    "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D",
                    "log",
                    "stream",
                    "--style",
                    "compact",
                    "--predicate",
                    'senderImagePath ENDSWITH "/First App"'
                    ' OR (processImagePath ENDSWITH "/First App"'
                    ' AND (senderImagePath ENDSWITH "-iphonesimulator.so"'
                    ' OR senderImagePath ENDSWITH "-iphonesimulator.dylib"))',
                ],
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                bufsize=1,
            ),
        ]
    )

    # Log stream monitoring was started
    run_command._stream_app_logs.assert_called_with(
        first_app_config,
        popen=log_stream_process,
        test_mode=True,
        clean_filter=macOS_log_clean_filter,
        clean_output=True,
        stop_func=mock.ANY,
        log_stream=True,
    )


@pytest.mark.usefixtures("sleep_zero")
def test_run_app_test_mode_with_passthrough(run_command, first_app_config, tmp_path):
    """An iOS App can be started in test mode with passthrough args."""
    # A valid target device will be selected.
    run_command.select_target_device = mock.MagicMock(
        return_value=("2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D", "13.2", "iPhone 11")
    )

    # Simulator is already booted
    run_command.get_device_state = mock.MagicMock(return_value=DeviceState.BOOTED)

    # Mock a process ID for the app
    run_command.tools.subprocess.check_output.return_value = (
        "com.example.first-app: 1234\n"
    )

    # Mock the uninstall, install, and log stream Popen processes
    uninstall_popen = mock.MagicMock(spec_set=subprocess.Popen)
    uninstall_popen.__enter__.return_value = uninstall_popen
    uninstall_popen.poll.side_effect = [None, None, 0]
    install_popen = mock.MagicMock(spec_set=subprocess.Popen)
    install_popen.__enter__.return_value = install_popen
    install_popen.poll.side_effect = [None, None, 0]
    log_stream_process = mock.MagicMock(spec_set=subprocess.Popen)
    run_command.tools.subprocess.Popen.side_effect = [
        uninstall_popen,
        install_popen,
        log_stream_process,
    ]

    # Run the app with args.
    run_command.run_app(
        first_app_config,
        test_mode=True,
        passthrough=["foo", "--bar"],
    )

    # Sleep four times for uninstall/install and once for log stream start
    assert time.sleep.call_count == 4 + 1

    # Open command was not called
    run_command.tools.subprocess.run.assert_not_called()

    # Launch the new app
    run_command.tools.subprocess.check_output.assert_called_once_with(
        [
            "xcrun",
            "simctl",
            "launch",
            "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D",
            "com.example.first-app",
            "foo",
            "--bar",
        ],
    )

    # Start the uninstall, install, and  log stream
    run_command.tools.subprocess.Popen.assert_has_calls(
        [
            # Uninstall the old app
            mock.call(
                [
                    "xcrun",
                    "simctl",
                    "uninstall",
                    "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D",
                    "com.example.first-app",
                ],
            ),
            # Install the new app
            mock.call(
                [
                    "xcrun",
                    "simctl",
                    "install",
                    "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D",
                    tmp_path
                    / "base_path/build/first-app/ios/xcode/build/Debug-iphonesimulator/First App.app",
                ],
            ),
            mock.call(
                [
                    "xcrun",
                    "simctl",
                    "spawn",
                    "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D",
                    "log",
                    "stream",
                    "--style",
                    "compact",
                    "--predicate",
                    'senderImagePath ENDSWITH "/First App"'
                    ' OR (processImagePath ENDSWITH "/First App"'
                    ' AND (senderImagePath ENDSWITH "-iphonesimulator.so"'
                    ' OR senderImagePath ENDSWITH "-iphonesimulator.dylib"))',
                ],
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                bufsize=1,
            ),
        ]
    )

    # Log stream monitoring was started
    run_command._stream_app_logs.assert_called_with(
        first_app_config,
        popen=log_stream_process,
        test_mode=True,
        clean_filter=macOS_log_clean_filter,
        clean_output=True,
        stop_func=mock.ANY,
        log_stream=True,
    )



import pytest


@pytest.mark.parametrize(
    "name",
    [
        # Various forms of capitalization and alphanumeric
        "Hello World",
        "helloworld",
        "helloWorld",
        "hello42world",
        "42helloworld",
        # Names that include punctuation
        "hello_world",
        "hello-world",
        "_helloworld",
        "/helloworld",
        "Hello / World!",
        # Internationalized names that can be unicode-simplified
        "Hallo Vögel",
        "Bonjour Garçon",
        # Internationalized names that cannot be unicode-simplified
        "你好 世界!",
    ],
)
def test_valid_formal_name(new_command, name):
    """Test that valid formal names are accepted."""
    assert new_command.validate_formal_name(name)


@pytest.mark.parametrize(
    "name",
    [
        "",  # Empty
        " ",  # Just a space
        "\t",  # Other whitespace characters
        "/",  # Just a slash
        "'",
        "\\",
        "/'\\",  # Multiple invalid characters
    ],
)
def test_invalid_formal_name(new_command, name, tmp_path):
    """Test that invalid app names are rejected."""
    (tmp_path / "existing").mkdir()

    with pytest.raises(ValueError):
        new_command.validate_formal_name(name)



import os
from unittest import mock

import pytest

from briefcase.console import Console


@pytest.fixture
def console(monkeypatch) -> Console:
    console = Console()
    console.input = mock.MagicMock(spec_set=input)
    # default console is always interactive
    monkeypatch.setattr(os, "isatty", lambda _: True)
    return console


@pytest.fixture
def disabled_console() -> Console:
    console = Console(enabled=False)
    console.input = mock.MagicMock(spec_set=input)
    return console


@pytest.fixture
def non_interactive_console(console, monkeypatch) -> Console:
    monkeypatch.setattr(os, "isatty", lambda _: False)
    yield console



import subprocess
import sys

import pytest

from briefcase.exceptions import BriefcaseCommandError

from ....utils import create_file


def create_emulator(root_path):
    # Create `emulator` within `root_path`.
    if sys.platform == "win32":
        emulator_bin = "emulator.exe"
    else:
        emulator_bin = "emulator"

    create_file(root_path / "emulator" / emulator_bin, "The Emulator", chmod=0o755)


def test_succeeds_immediately_if_emulator_installed(mock_tools, android_sdk):
    """`verify_emulator()` exits early if the emulator exists in its root_path."""
    # Create `emulator` within `root_path`.
    create_emulator(android_sdk.root_path)

    # Also create the platforms folder
    (android_sdk.root_path / "platforms").mkdir(parents=True)

    android_sdk.verify_emulator()

    # Platforms folder still exists
    assert (android_sdk.root_path / "platforms").exists()

    # No extra calls made
    mock_tools.subprocess.run.assert_not_called()
    mock_tools.file.download.assert_not_called()


def test_creates_platforms_folder(mock_tools, android_sdk):
    """If the platforms folder doesn't exist, it is created."""
    # Create `emulator` within `root_path`.
    create_emulator(android_sdk.root_path)

    # Verify the emulator. This should create the missing "platforms" folder.
    android_sdk.verify_emulator()

    # Platforms folder now exists
    assert (android_sdk.root_path / "platforms").exists()

    # No extra calls made
    mock_tools.subprocess.run.assert_not_called()
    mock_tools.file.download.assert_not_called()


def test_installs_android_emulator(mock_tools, android_sdk):
    """The emulator tools will be installed if needed."""
    android_sdk.verify_emulator()

    # Platforms folder now exists
    assert (android_sdk.root_path / "platforms").exists()

    mock_tools.subprocess.run.assert_called_once_with(
        [
            android_sdk.sdkmanager_path,
            "platform-tools",
            "emulator",
        ],
        env=android_sdk.env,
        check=True,
        stream_output=False,
    )


def test_partial_android_emulator_install(mock_tools, android_sdk):
    """If the Android emulator is only partially installed, it's not detected."""
    # Create the emulator *directory*, but not the actual binary.
    (android_sdk.root_path / "emulator").mkdir(parents=True)

    android_sdk.verify_emulator()

    # Platforms folder now exists
    assert (android_sdk.root_path / "platforms").exists()

    mock_tools.subprocess.run.assert_called_once_with(
        [
            android_sdk.sdkmanager_path,
            "platform-tools",
            "emulator",
        ],
        env=android_sdk.env,
        check=True,
        stream_output=False,
    )


def test_install_problems_are_reported(mock_tools, android_sdk):
    """If the sdkmanager fails to properly install the Android emulator, an exception is
    raised."""
    # Configure `subprocess` module to crash as though it were a sad sdkmanager.
    mock_tools.subprocess.run.side_effect = subprocess.CalledProcessError(
        returncode=1,
        cmd=["ignored"],
    )
    with pytest.raises(BriefcaseCommandError):
        android_sdk.verify_emulator()



from __future__ import annotations

import re
import subprocess
from abc import abstractmethod
from contextlib import suppress

from briefcase.config import AppConfig
from briefcase.exceptions import BriefcaseCommandError, BriefcaseTestSuiteFailure
from briefcase.integrations.subprocess import StopStreaming

from .base import BaseCommand, full_options


class LogFilter:
    DEFAULT_EXIT_REGEX = r"^>>>>>>>>>> EXIT (?P<returncode>.*) <<<<<<<<<<$"

    def __init__(
        self,
        log_popen,
        clean_filter,
        clean_output,
        exit_filter,
    ):
        """Create a filter for a log stream.

        :param log_popen: The Popen object for the stream producing the logs.
        :param clean_filter: A function that will filter a line of logs, returning a
            "clean" line without any log system preamble.
        :param clean_output: Should the output displayed to the user be the "clean"
            output? (Default: True).
        :param exit_filter: A function that will operate on a string containing the last
            10 lines of "clean" (i.e., preamble filtered) logs, returning the integer
            exit status of the process if an exit condition has been detected, or None
            if the log stream should continue.
        """
        self.log_popen = log_popen
        self.returncode = None
        self.clean_filter = clean_filter
        self.clean_output = clean_output

        self.recent_history = []
        self.exit_filter = exit_filter

    def __call__(self, line):
        """Filter a single line of a log.

        :param line: A single line of raw system log content, including the newline.
        """
        # Compute the clean line
        if self.clean_filter:
            filtered = self.clean_filter(line)

            # If the clean filter says the line can be dumped, return without yielding
            if filtered is None:
                return

            # If there's a cleaned line, we can determine if it should be included in analysis
            clean_line, included = filtered
        else:
            # If we don't perform cleaning, we assume all content is potentially
            # Python, and should be included in analysis
            clean_line = line
            included = True

        # If we're not using clean output, use the raw line for display.
        if self.clean_output:
            display_line = clean_line
        else:
            display_line = line

        # If the line is Python content, append the new line, then clip the
        # recent history to the most recent 10 clean lines, and build a single
        # string that is the tail of the recent clean lines.
        if included:
            self.recent_history.append(clean_line)
            self.recent_history = self.recent_history[-10:]
            tail = "\n".join(self.recent_history)

            # Look for the exit condition in the tail
            if self.exit_filter:
                self.returncode = self.exit_filter(tail)
                if self.returncode is not None:
                    # This returncode is captured from output from the app and does not
                    # necessarily mean the app has already exited. However, once
                    # StopStreaming is raised, the output streamer sends a signal to
                    # the app to exit immediately and that may result in the app
                    # exiting with a non-zero returncode. Therefore, wait for the app
                    # to close normally before raising StopStreaming.
                    with suppress(subprocess.TimeoutExpired):
                        self.log_popen.wait(timeout=3)
                    raise StopStreaming()

        # Return the display line
        yield display_line

    @staticmethod
    def test_filter(pattern):
        """A factory method for producing filter functions.

        :param pattern: The multiline regex pattern that identifies content of interest
            in a log (e.g., success/failure conditions)
        :returns: A log filter function that returns True if the pattern was found
        """

        def filter_func(recent):
            match = filter_func.regex.search(recent)
            if match:
                try:
                    return int(match.groupdict()["returncode"])
                except KeyError:
                    # No returncode group in regex
                    return -998
                except ValueError:
                    # Non-integer return code content
                    return -999
            return None

        # Annotate the function with the regex that will be used in the function.
        filter_func.regex = re.compile(pattern, re.MULTILINE)
        return filter_func


class RunAppMixin:
    """A mixin that captures the logic of starting an app and streaming the app logs."""

    allows_passthrough = True

    def _stream_app_logs(
        self,
        app: AppConfig,
        popen,
        test_mode=False,
        clean_filter=None,
        clean_output=False,
        stop_func=lambda: False,
        log_stream=False,
    ):
        """Stream the application's logs, monitoring for exit conditions.

        Catches and cleans up after any Ctrl-C interrupts.

        :param app: The app to be launched
        :param popen: The Popen object for the stream we are monitoring; this Popen
            process will be closed after log streaming completes.
        :param test_mode: Are we launching in test mode?
        :param clean_filter: The log cleaning filter to use; see ``LogFilter``
            for details.
        :param clean_output: Should the cleaned output be presented to the user?
        :param stop_func: (Optional) A function that will be invoked to determine
            if the log stream should be terminated.
        :param log_stream: Is this a log stream, rather than a literal app stream?
            On some platforms (especially mobile), we monitor a log stream,
            rather that the output of the app itself. If this case, the cleanup
            process is different, as the reported exit status of the popen object
            is of the log, not the app itself.
        """
        try:
            exit_filter = LogFilter.test_filter(
                getattr(app, "exit_regex", LogFilter.DEFAULT_EXIT_REGEX)
            )

            log_filter = LogFilter(
                popen,
                clean_filter=clean_filter,
                clean_output=clean_output,
                exit_filter=exit_filter,
            )

            # Start streaming logs for the app.
            self.logger.info("=" * 75)
            with popen:
                self.tools.subprocess.stream_output(
                    label="log stream" if log_stream else app.app_name,
                    popen_process=popen,
                    stop_func=stop_func,
                    filter_func=log_filter,
                )

            # If we're in test mode, and log streaming ends,
            # check for the status of the test suite.
            if test_mode:
                if log_filter.returncode == 0:
                    self.logger.info("Test suite passed!", prefix=app.app_name)
                else:
                    if log_filter.returncode is None:
                        raise BriefcaseCommandError(
                            "Test suite didn't report a result."
                        )
                    else:
                        self.logger.error("Test suite failed!", prefix=app.app_name)
                        raise BriefcaseTestSuiteFailure()
            elif log_stream:
                # If we're monitoring a log stream, and the log stream reported a
                # non-zero exit code, surface that error to the user.
                if log_filter.returncode is not None and log_filter.returncode != 0:
                    raise BriefcaseCommandError(f"Problem running app {app.app_name}.")
            else:
                # If we're monitoring an actual app (not just a log stream),
                # and the app didn't exit cleanly, surface the error to the user.
                if popen.poll() != 0:
                    raise BriefcaseCommandError(f"Problem running app {app.app_name}.")

        except KeyboardInterrupt:
            pass  # Catch CTRL-C to exit normally


class RunCommand(RunAppMixin, BaseCommand):
    command = "run"
    description = "Run an app."

    def add_options(self, parser):
        parser.add_argument(
            "-a",
            "--app",
            dest="appname",
            help="The app to run",
        )

        self._add_update_options(parser, context_label=" before running")
        self._add_test_options(parser, context_label="Run")

    def _prepare_app_kwargs(self, app: AppConfig, test_mode: bool):
        """Prepare the kwargs for running an app as a log stream.

        This won't be used by every backend; but it's a sufficiently common default that
        it's been factored out.

        :param app: The app to be launched
        :param test_mode: Are we launching in test mode?
        :returns: A dictionary of additional arguments to pass to the Popen
        """
        args = {}
        env = {}

        # If we're in debug mode, put BRIEFCASE_DEBUG into the environment
        if self.logger.is_debug:
            env["BRIEFCASE_DEBUG"] = "1"

        if test_mode:
            # In test mode, set a BRIEFCASE_MAIN_MODULE environment variable
            # to override the module at startup
            env["BRIEFCASE_MAIN_MODULE"] = app.main_module(test_mode)
            self.logger.info("Starting test_suite...", prefix=app.app_name)
        else:
            self.logger.info("Starting app...", prefix=app.app_name)

        # If we need any environment variables, add them to the arguments.
        if env:
            args["env"] = env

        return args

    @abstractmethod
    def run_app(self, app: AppConfig, **options) -> dict | None:
        """Start an application.

        :param app: The application to start
        """

    def __call__(
        self,
        appname: str | None = None,
        update: bool = False,
        update_requirements: bool = False,
        update_resources: bool = False,
        update_support: bool = False,
        update_stub: bool = False,
        no_update: bool = False,
        test_mode: bool = False,
        passthrough: list[str] | None = None,
        **options,
    ) -> dict | None:
        # Which app should we run? If there's only one defined
        # in pyproject.toml, then we can use it as a default;
        # otherwise look for a -a/--app option.
        if len(self.apps) == 1:
            app = list(self.apps.values())[0]
        elif appname:
            try:
                app = self.apps[appname]
            except KeyError as e:
                raise BriefcaseCommandError(
                    f"Project doesn't define an application named '{appname}'"
                ) from e
        else:
            raise BriefcaseCommandError(
                "Project specifies more than one application; use --app to specify which one to start."
            )

        # Confirm host compatibility, that all required tools are available,
        # and that the app configuration is finalized.
        self.finalize(app)

        template_file = self.bundle_path(app)
        exec_file = self.binary_executable_path(app)
        if (
            (not template_file.exists())  # App hasn't been created
            or update  # An explicit update has been requested
            or update_requirements  # An explicit update of requirements has been requested
            or update_resources  # An explicit update of resources has been requested
            or update_support  # An explicit update of support files has been requested
            or update_stub  # An explicit update of the stub binary has been requested
            or (not exec_file.exists())  # Executable binary doesn't exist yet
            or (
                test_mode and not no_update
            )  # Test mode, but updates have not been disabled
        ):
            state = self.build_command(
                app,
                update=update,
                update_requirements=update_requirements,
                update_resources=update_resources,
                update_support=update_support,
                update_stub=update_stub,
                no_update=no_update,
                test_mode=test_mode,
                **options,
            )
        else:
            state = None

        self.verify_app(app)

        state = self.run_app(
            app,
            test_mode=test_mode,
            passthrough=[] if passthrough is None else passthrough,
            **full_options(state, options),
        )

        return state



import pytest

from briefcase.exceptions import MissingToolError, NetworkFailure
from tests.integrations.linuxdeploy.utils import side_effect_create_mock_appimage


def test_upgrade_exists(linuxdeploy, mock_tools, tmp_path):
    """If linuxdeploy already exists, upgrading deletes first."""
    appimage_path = tmp_path / "tools/linuxdeploy-i386.AppImage"

    # Mock the existence of an install
    appimage_path.touch()

    # Mock a successful download
    mock_tools.file.download.side_effect = side_effect_create_mock_appimage(
        appimage_path
    )

    # Create a linuxdeploy wrapper, then upgrade it
    linuxdeploy.upgrade()

    # The mock file should exist as the upgraded version
    assert appimage_path.exists()

    # A download is invoked
    mock_tools.file.download.assert_called_with(
        url="https://github.com/linuxdeploy/linuxdeploy/releases/download/continuous/linuxdeploy-i386.AppImage",
        download_path=tmp_path / "tools",
        role="linuxdeploy",
    )
    # The downloaded file will be made executable
    mock_tools.os.chmod.assert_called_with(appimage_path, 0o755)


def test_upgrade_does_not_exist(linuxdeploy, mock_tools):
    """If linuxdeploy doesn't already exist, upgrading is an error."""
    # Create a linuxdeploy wrapper, then upgrade it
    with pytest.raises(MissingToolError):
        linuxdeploy.upgrade()

    # The tool wasn't already installed, so an error is raised.
    assert mock_tools.file.download.call_count == 0


def test_upgrade_linuxdeploy_download_failure(linuxdeploy, mock_tools, tmp_path):
    """If linuxdeploy doesn't exist, but a download failure occurs, an error is
    raised."""
    appimage_path = tmp_path / "tools/linuxdeploy-i386.AppImage"

    # Mock the existence of an install
    appimage_path.touch()

    mock_tools.file.download.side_effect = NetworkFailure("mock")

    # Updated the linuxdeploy wrapper; the upgrade will fail
    with pytest.raises(NetworkFailure, match="Unable to mock"):
        linuxdeploy.upgrade()

    # The mock file will be deleted
    assert not appimage_path.exists()

    # A download was invoked
    mock_tools.file.download.assert_called_with(
        url="https://github.com/linuxdeploy/linuxdeploy/releases/download/continuous/linuxdeploy-i386.AppImage",
        download_path=tmp_path / "tools",
        role="linuxdeploy",
    )



from __future__ import annotations

import logging
import operator
import os
import platform
import re
import shutil
import sys
import textwrap
import time
import traceback
from contextlib import contextmanager
from datetime import datetime
from enum import IntEnum
from pathlib import Path
from typing import Callable, Iterable

from rich.console import Console as RichConsole
from rich.control import strip_control_codes
from rich.highlighter import RegexHighlighter
from rich.markup import escape
from rich.progress import (
    BarColumn,
    Progress,
    SpinnerColumn,
    TextColumn,
    TimeRemainingColumn,
)
from rich.traceback import Trace, Traceback

from briefcase import __version__

# Max width for printing to console; matches argparse's default width
MAX_TEXT_WIDTH = max(min(shutil.get_terminal_size().columns, 80) - 2, 20)

# Regex to identify settings likely to contain sensitive information
SENSITIVE_SETTING_RE = re.compile(r"API|TOKEN|KEY|SECRET|PASS|SIGNATURE", flags=re.I)

# 7-bit C1 ANSI escape sequences
ANSI_ESC_SEQ_RE_DEF = r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])"
ANSI_ESCAPE_RE = re.compile(ANSI_ESC_SEQ_RE_DEF)


class InputDisabled(Exception):
    def __init__(self):
        super().__init__(
            "Input is disabled; cannot request user input without a default"
        )


def sanitize_text(text: str) -> str:
    """Remove control codes and ANSI escape sequences from a line of text.

    This is useful for extracting the plain text from the output of third party tools
    that may be including markup for display in the console.
    """
    return ANSI_ESCAPE_RE.sub("", strip_control_codes(text))


class RichConsoleHighlighter(RegexHighlighter):
    """Custom Rich highlighter for printing to the console.

    This highlighter limits text highlighting to only URLs.

    By default, Rich applies several highlighting rules to anything it prints for
    concepts like UUIDs, markup, IP addresses, numbers, etc. All these colorful
    additions to the text can become overbearing and distracting given much of the
    output does not benefit from coloring.

    This defines a visually simpler style that highlights URLs in a way that is vaguely
    consistent with a default HTML stylesheet, but otherwise renders content as-is.
    """

    base_style = "repr."
    highlights = [
        r"(?P<url>(file|https|http|ws|wss)://[-0-9a-zA-Z$_+!`(),.?/;:&=%#~]*)"
    ]


class Printer:
    def __init__(self, log_width=180):
        """Interface for printing and managing output to the console and/or log.

        The default width is wide enough to render the output of ``sdkmanager
        --list_installed`` in the log file without line wrapping.

        :param log_width: The width at which content should be wrapped in the log file
        """
        self.log_width = log_width

        # A wrapper around the console
        self.console = RichConsole(
            highlighter=RichConsoleHighlighter(),
            emoji=False,
            soft_wrap=True,
        )

        # Rich only records what's being logged if it is actually written somewhere;
        # writing to /dev/null allows Rich to do so without needing to print the
        # logs in the console or save them to file before it is known a file is wanted.
        self.dev_null = open(os.devnull, "w", encoding="utf-8", errors="ignore")
        self.log = RichConsole(
            file=self.dev_null,
            record=True,
            width=self.log_width,
            force_interactive=False,
            force_terminal=False,
            no_color=True,
            color_system=None,
            markup=False,
            emoji=False,
            highlight=False,
            soft_wrap=True,
        )

    def __del__(self):
        self.dev_null.close()

    def __call__(self, *messages, stack_offset=5, show=True, **kwargs):
        """Entry point for all printing to the console and the log.

        The log records all content that is printed whether it is shown in the console
        or not (e.g. debug output). When execution completes, the log is conditionally
        exported and saved to a file.

        :param messages: content to print and/or log
        :param show: True (default) to print and log messages; False to only log.
        :param stack_offset: number of levels up the stack where logging was invoked.
            This tells Rich the number of levels to recurse up the stack to find the
            filename to put in the right column of the Rich log. Defaults to 5 since
            most uses are 5 levels deep from the actual logging.
        """
        if show:
            self.to_console(*messages, **kwargs)
        self.to_log(*messages, stack_offset=stack_offset, **kwargs)

    def to_console(self, *messages, **kwargs):
        """Write only to the console and skip writing to the log."""
        self.console.print(*messages, **kwargs)

    def to_log(self, *messages, stack_offset=5, **kwargs):
        """Write only to the log and skip writing to the console."""
        self.log.log(
            *map(sanitize_text, messages), _stack_offset=stack_offset, **kwargs
        )

    def export_log(self):
        """Export the text of the entire log; the log is also cleared."""
        return self.log.export_text()


class RichLoggingStream:
    """Stream for logging.StreamHandler that prints to console via debug logging."""

    def __init__(self, logger: Log):
        self.logger = logger

    def write(self, msg: str) -> None:
        self.logger.debug(msg)


class RichLoggingHandler(logging.StreamHandler):
    """A debug handler for third party tools using stdlib logging."""

    def __init__(self, stream: RichLoggingStream):
        super().__init__(stream=stream)
        self.setLevel(logging.DEBUG)
        self.setFormatter(logging.Formatter("%(levelname)s %(name)s: %(message)s"))


class LogLevel(IntEnum):
    INFO = 0
    VERBOSE = 1
    DEBUG = 2
    DEEP_DEBUG = 3


class Log:
    """Manage logging output driven by verbosity flags."""

    # subdirectory of command.base_path to store log files
    LOG_DIR = "logs"

    def __init__(
        self,
        printer: Printer | None = None,
        verbosity: LogLevel = LogLevel.INFO,
    ):
        self.print = Printer() if printer is None else printer
        # --verbosity flag: see LogLevel for valid values
        self.verbosity = verbosity
        # --log flag to force logfile creation
        self.save_log = False
        # flag set by exceptions to skip writing the log; save_log takes precedence.
        self.skip_log = False
        # Rich stacktraces of exceptions for logging to file.
        # A list of tuples containing a label for the thread context, and the Trace object
        self.stacktraces: list[tuple[str, Trace]] = []
        # functions to run for additional logging if creating a logfile
        self.log_file_extras: list[Callable[[], object]] = []
        # The current context for the log
        self._context = ""

    @contextmanager
    def context(self, context: str):
        """Wrap a collection of output in a logging context.

        A logging context is a prefix on every logging line. It is used when a set of
        commands (and output) is being run in a very specific way that needs to be
        highlighted, such as running a command in a Docker container.

        :param context: The name of the context to enter. This *must* be simple text,
            with no markup or other special characters.
        """
        self.info()
        self.info(f"Entering {context} context...")
        old_context = self._context
        self._context = f"{context}| "
        self.info("-" * (72 - len(context)))
        try:
            yield
        finally:
            self.info("-" * (72 - len(context)))
            self._context = old_context
            self.info(f"Leaving {context} context.")
            self.info()

    def _log(
        self,
        preface="",
        prefix="",
        message="",
        show=True,
        markup=False,
        style=None,
    ):
        """Funnel to log all messages.

        :param preface: value to be printed on the far left of the message and symbolic
            of the type of message being printed.
        :param prefix: text prepended to the message wrapped in brackets and will be
            presented as dimmer compared to the styling of the message text.
        :param message: text to log; can contain Rich tags if markup=True.
        :param show: True to print message to console; False to not print it. This is to
            allow logs to be saved to a file without printing them.
        :param markup: whether to interpret Rich markup in the prefix, preface, and
            message; if True, all text must already be escaped; defaults False.
        :param style: Rich style to apply to everything printed for message.
        """
        if not message:
            # When a message is not provided, only output the context;
            # This type of call is just clearing some vertical space.
            self.print(self._context, show=show)
        else:
            if prefix:
                # insert vertical space before for all messages with a prefix
                self.print(self._context, show=show)
                if not markup:
                    preface, prefix, message = (
                        escape(text) for text in (preface, prefix, message)
                    )
                prefix = f"[dim]\\[{prefix}][/dim] "
                markup = True
            for line in message.splitlines():
                self.print(
                    f"{self._context}{preface}{prefix}{line}",
                    show=show,
                    markup=markup,
                    style=style,
                )

    def debug(self, message="", *, preface="", prefix="", markup=False):
        """Log messages at debug level; included if verbosity >= 2."""
        self._log(
            preface=preface,
            prefix=prefix,
            message=message,
            show=self.is_debug,
            markup=markup,
            style="dim",
        )

    def verbose(self, message="", *, prefix="", markup=False):
        """Log message at verbose level; included if verbosity >= 1."""
        self._log(prefix=prefix, message=message, show=self.is_verbose, markup=markup)

    def info(self, message="", *, prefix="", markup=False):
        """Log message at info level; always included in output."""
        self._log(prefix=prefix, message=message, markup=markup)

    def warning(self, message="", *, prefix="", markup=False):
        """Log message at warning level; always included in output."""
        self._log(prefix=prefix, message=message, markup=markup, style="bold yellow")

    def error(self, message="", *, prefix="", markup=False):
        """Log message at error level; always included in output."""
        self._log(prefix=prefix, message=message, markup=markup, style="bold red")

    @property
    def verbosity(self) -> LogLevel:
        return self._verbosity

    @verbosity.setter
    def verbosity(self, value: int | LogLevel):
        """Clamp verbosity between valid log levels."""
        self._verbosity = LogLevel(max(min(LogLevel), min(value, max(LogLevel))))

    @property
    def is_verbose(self):
        """Is verbose logging enabled via the -v flag?

        This includes Briefcase logging that isn't relevant during normal operation.
        """
        return self.verbosity >= LogLevel.VERBOSE

    @property
    def is_debug(self):
        """Is debug logging enabled via the -vv flag?

        This includes debug logging from third party tools.
        """
        return self.verbosity >= LogLevel.DEBUG

    @property
    def is_deep_debug(self):
        """Is deep debug logging enabled via the -vvv flag?

        This includes debug logging from third party tools.
        """
        return self.verbosity >= LogLevel.DEEP_DEBUG

    def configure_stdlib_logging(self, logger_name: str):
        """Configure stdlib logging for a third party tool to log through Rich.

        When a third party tool written in Python uses the stdlib logging for their
        logging, it may provide an abstraction to enable the logging in the console or
        may require a handler to be added externally. Either way, the default handler
        to write to the console, i.e. logging.StreamHandler, will bypass Rich logging
        and therefore not be included in the logfile. To avoid this issue, this will
        add a handler that specifically writes to the console through Rich.

        :param logger_name: Name of the logger the third party tool uses. Typically,
            this is the package name or path to a submodule for the package. Since
            logging uses namespaces for logging, enabling DEBUG logging for `git` will
            also enable it for others like `git.cmd`, `git.remote`, `git.config`, etc.
        """
        if not self.is_deep_debug:
            return

        logger = logging.getLogger(logger_name)

        # do not add another rich handler if one already exists
        if not any(isinstance(h, RichLoggingHandler) for h in logger.handlers):
            logger.addHandler(RichLoggingHandler(stream=RichLoggingStream(logger=self)))
            logger.setLevel(logging.DEBUG)

    def capture_stacktrace(self, label="Main thread"):
        """Preserve Rich stacktrace from exception while in except block.

        :param label: An identifying label for the thread that has raised the
            stacktrace. Defaults to the main thread.
        """
        exc_info = sys.exc_info()
        try:
            self.skip_log = exc_info[1].skip_logfile
        except AttributeError:
            pass

        self.stacktraces.append((label, Traceback.extract(*exc_info, show_locals=True)))

    def add_log_file_extra(self, func: Callable[[], object]):
        """Register a function to be called in the event that a log file is written.

        This can be used to provide additional debugging information which is too
        expensive to gather pre-emptively.
        """
        self.log_file_extras.append(func)

    def save_log_to_file(self, command):
        """Save the current application log to file."""
        # A log file is always written if a Command ran and `--log` was provided.
        # If `--log` was not provided, then the log is written when an exception
        # occurred and the exception was not explicitly configured to skip the log.
        if not (
            command and (self.save_log or (self.stacktraces and not self.skip_log))
        ):
            return

        with command.input.wait_bar("Saving log...", transient=True):
            self.print.to_console()
            log_filename = f"briefcase.{datetime.now().strftime('%Y_%m_%d-%H_%M_%S')}.{command.command}.log"
            log_filepath = command.base_path / self.LOG_DIR / log_filename
            try:
                log_filepath.parent.mkdir(parents=True, exist_ok=True)
                with open(
                    log_filepath, "w", encoding="utf-8", errors="backslashreplace"
                ) as log_file:
                    log_file.write(self._build_log(command))
            except OSError as e:
                self.error(f"Failed to save log to {log_filepath}: {e}")
            else:
                self.warning(f"Log saved to {log_filepath}")
            self.print.to_console()

    def _build_log(self, command) -> str:
        """Accumulate all information to include in the log file."""
        # Add the exception stacktraces to end of log if any were captured
        if self.stacktraces:
            # using print.log.print() instead of print.to_log() to avoid
            # timestamp and code location inclusion for the stacktrace box.
            for thread, stacktrace in self.stacktraces:
                self.print.log.print(
                    f"{thread} traceback:",
                    Traceback(
                        trace=stacktrace,
                        width=self.print.log_width,
                        show_locals=True,
                    ),
                    new_line_start=True,
                )

        # Retrieve additional logging added by the Command
        if self.log_file_extras:
            with command.input.wait_bar(
                "Collecting extra information for log...",
                transient=True,
            ):
                self.debug()
                self.debug("Extra information:")
                for func in self.log_file_extras:
                    try:
                        func()
                    except Exception:
                        self.error(traceback.format_exc())

        # Capture env vars removing any potentially sensitive information
        sanitized_env_vars = "\n".join(
            f"\t{env_var}={value if not SENSITIVE_SETTING_RE.search(env_var) else '********************'}"
            for env_var, value in sorted(command.tools.os.environ.items())
        )

        # Capture pyproject.toml if one exists in the current directory
        try:
            with open(Path.cwd() / "pyproject.toml", encoding="utf-8") as f:
                pyproject_toml = f.read().strip()
        except OSError as e:
            pyproject_toml = str(e)

        # Build log with buffered log from Rich
        uname = platform.uname()
        return (
            f"Date/Time:       {datetime.now().strftime('%Y-%m-%d %H:%M:%S %Z')}\n"
            f"Command line:    {' '.join(sys.argv)}\n"
            "\n"
            f"OS Release:      {uname.system} {uname.release}\n"
            f"OS Version:      {uname.version}\n"
            f"Architecture:    {uname.machine}\n"
            f"Platform:        {platform.platform(aliased=True)}\n"
            "\n"
            f"Python exe:      {sys.executable}\n"
            # replace line breaks with spaces (use chr(10) since '\n' isn't allowed in f-strings...)
            f"Python version:  {sys.version.replace(chr(10), ' ')}\n"
            # sys.real_prefix was used in older versions of virtualenv.
            # sys.base_prefix is always the python exe's original site-specific directory (e.g. /usr/local).
            # sys.prefix is updated (from base_prefix's value) to the virtual env's site-specific directory.
            f"Virtual env:     {hasattr(sys, 'real_prefix') or sys.base_prefix != sys.prefix}\n"
            # for conda, prefix and base_prefix are likely the same but contain a conda-meta dir.
            f"Conda env:       {(Path(sys.prefix) / 'conda-meta').exists()}\n"
            "\n"
            f"Briefcase:       {__version__}\n"
            f"Target platform: {command.platform}\n"
            f"Target format:   {command.output_format}\n"
            "\n"
            "Environment Variables:\n"
            f"{sanitized_env_vars}\n"
            "\n"
            "pyproject.toml:\n"
            f"{pyproject_toml}\n"
            "\n"
            "Briefcase Log:\n"
            f"{self.print.export_log()}"
        )


class NotDeadYet:
    # I’m getting better! No you’re not, you’ll be stone dead in a minute.

    def __init__(self, printer: Printer):
        """A keep-alive spinner for long-running processes without console output.

        Returned by the Wait Bar's context manager but can be used independently. Use in
        a loop that calls update() each iteration. A keep-alive message will be printed
        every 10 seconds.
        """
        self.printer: Printer = printer
        self.interval_sec: int = 10
        self.message: str = "... still waiting"
        self.ready_time: float = 0.0

        self.reset()  # initialize

    def update(self):
        """Write keep-alive message if the periodic interval has elapsed."""
        if self.ready_time < time.time():
            self.printer(self.message)
            self.reset()

    def reset(self):
        """Initialize periodic interval to now; next message prints after interval."""
        self.ready_time = time.time() + self.interval_sec


class Console:
    def __init__(self, printer: Printer | None = None, enabled: bool = True):
        self.enabled = enabled
        self.print = Printer() if printer is None else printer
        # Use Rich's input() to read from user
        self.input = self.print.console.input
        self._wait_bar: Progress | None = None
        # Signal that Rich is dynamically controlling the console output. Therefore,
        # all output must be printed to the screen by Rich to prevent corruption of
        # dynamic elements like the Wait Bar.
        self.is_console_controlled = False

    @property
    def is_interactive(self):
        """Returns interactivity mode based on the presence of a tty for stdout.

        In a non-interactive session, Rich's dynamic elements like progress bars should
        be disabled. By default, Rich will not render dynamic elements if a tty is not
        available. However, environment variables like FORCE_COLOR can override this
        behavior in Rich to not only show ANSI color and style but also dynamic
        elements. In a CI environment with this configuration, a long-lived progress bar
        can dramatically compromise the quality of logged output. So, dynamic elements
        should be specifically disabled in non-interactive sessions.
        """
        # `sys.__stdout__` is used because Rich captures and redirects `sys.stdout`
        return os.isatty(sys.__stdout__.fileno())

    @property
    def is_color_enabled(self):
        """Is the underlying Rich console using color?

        Rich can be explicitly configured to not use color at Console initialization or
        the NO_COLOR environment variable; alternatively, the derived color system for
        the terminal is influenced by attributes of the platform as well as FORCE_COLOR.
        """
        # no_color has precedence since color_system can be set even if color is disabled
        if self.print.console.no_color:
            return False
        else:
            return self.print.console.color_system is not None

    def progress_bar(self):
        """Returns a progress bar as a context manager."""
        return Progress(
            TextColumn("  "),
            SpinnerColumn("line", speed=1.5, style="default"),
            BarColumn(bar_width=50),
            TextColumn("{task.percentage:>3.1f}%", style="default"),
            TextColumn("•", style="default"),
            TimeRemainingColumn(compact=True, elapsed_when_finished=True),
            disable=not self.is_interactive,
            console=self.print.console,
        )

    @contextmanager
    def wait_bar(
        self,
        message: str = "",
        done_message: str = "done",
        *,
        transient: bool = False,
        markup: bool = False,
    ) -> NotDeadYet:
        """Activates the Wait Bar as a context manager.

        If the Wait Bar is already active, then its message is updated for the new
        context. Once the new context is complete, the previous Wait Bar message
        is restored.

        :param message: text explaining what is being awaited; should end in '...'
        :param done_message: text appended to the message after exiting
        :param transient: if True, remove bar and message from screen after exiting;
            if False (default), the message will remain on the screen without pulsing bar.
        :param markup: whether to interpret Rich styling markup in the message; if True,
            the message must already be escaped; defaults False.
        :returns:  Keep-alive spinner to notify user Briefcase is still waiting
        """
        is_wait_bar_disabled = not self.is_interactive
        show_outcome_message = message and (is_wait_bar_disabled or not transient)

        if self._wait_bar is None:
            self._wait_bar = Progress(
                TextColumn("    "),
                BarColumn(bar_width=20, style="black", pulse_style="white"),
                TextColumn("{task.fields[message]}"),
                transient=True,
                disable=is_wait_bar_disabled,
                console=self.print.console,
            )
            # start=False causes the progress bar to "pulse"
            # message=None is a sentinel the Wait Bar should be inactive
            self._wait_bar.add_task("", start=False, message=None)

        self.print(
            f"{message} started", markup=markup, show=message and is_wait_bar_disabled
        )

        self.is_console_controlled = True
        wait_bar_task = self._wait_bar.tasks[0]
        previous_message = wait_bar_task.fields["message"]
        self._wait_bar.update(wait_bar_task.id, message=message)

        try:
            self._wait_bar.start()
            yield NotDeadYet(printer=self.print)
        except BaseException as e:
            # capture BaseException so message is left on the screen even if user sends CTRL+C
            error_message = "aborted" if isinstance(e, KeyboardInterrupt) else "errored"
            self.print(
                f"{message} {error_message}", markup=markup, show=show_outcome_message
            )
            raise
        else:
            self.print(
                f"{message} {done_message}", markup=markup, show=show_outcome_message
            )
        finally:
            self._wait_bar.update(wait_bar_task.id, message=previous_message)
            # Deactivate the Wait Bar if returning to its initial state
            if previous_message is None:
                self._wait_bar.stop()
                self.is_console_controlled = False

    @contextmanager
    def release_console_control(self):
        """Context manager to remove console elements such as the Wait Bar.

        This is useful to temporarily release control of the console when, e.g., a
        process is interrupted or a user needs to be prompted. For instance, when batch
        scripts are interrupted by CTRL+C in cmd.exe, the user may be prompted to abort
        the script; so, the console cannot be controlled while such scripts run or the
        prompt may be hidden from the user.
        """
        # Preserve current console state
        is_output_controlled = self.is_console_controlled
        is_wait_bar_running = self._wait_bar and self._wait_bar.live.is_started

        # Stop any active dynamic console elements
        if is_wait_bar_running:
            self._wait_bar.stop()

        self.is_console_controlled = False
        try:
            yield
        finally:
            self.is_console_controlled = is_output_controlled
            # Restore previous console state
            if is_wait_bar_running:
                self._wait_bar.start()

    def textwrap(self, text: str, width: int = MAX_TEXT_WIDTH) -> str:
        """Wrap text to the console width, a default max width, or a specified width."""
        # textwrap isn't really designed to format text that already contains newlines.
        # So, instead, break the text by newlines and format each line individually.
        return "\n".join(
            "\n".join(textwrap.wrap(line, width)) for line in text.splitlines()
        )

    def prompt(self, *values, markup=False, **kwargs):
        """Print to the screen for soliciting user interaction if input enabled.

        :param values: strings to print as the user prompt
        :param markup: True if prompt contains Rich markup
        """
        if self.enabled:
            self.print(*values, markup=markup, stack_offset=4, **kwargs)

    def boolean_input(self, question: str, default: bool = False) -> bool:
        """Get a boolean input from user, in the form of y/n.

        The user might press "y" for true or "n" for false. If input is disabled,
        returns default. If input is disabled and default is *not* defined,
        InputDisabled is raised.

        :param question: A string message specifying the question to be answered by the
            user.
        :param default: (optional) Default response (True/False)
        :returns: True if the user selected "y", or False if they selected "n".
        """
        if default is None:
            yes_no = "y/n"
            default_text = None
        elif default:
            yes_no = "[Y/n]"
            default_text = "y"
        else:
            yes_no = "[y/N]"
            default_text = "n"

        prompt = f"{question} {yes_no}? "

        result = self.selection_input(
            prompt=prompt,
            choices=["y", "n"],
            default=default_text,
            error_message="Please enter Y or N",
            transform=lambda s: s.lower()[:1],
        )
        if result == "y":
            return True

        return False

    def selection_input(
        self,
        prompt: str,
        choices: Iterable[str],
        default: str | None = None,
        error_message: str = "Invalid Selection",
        transform: Callable[[str], str] | None = None,
    ) -> str:
        """Prompt the user to select an option from a list of choices.

        :param prompt: The text prompt to display
        :param choices: The list of available choices
        :param default: The default choice to select. If None,
        :param error_message: The error message to display to the user.
        :param transform: The text transform to apply to any user input before
            performing any validity checks.
        """
        while True:
            result = self.text_input(prompt, default)

            if transform is not None and result is not None:
                result = transform(result)

            if result in choices:
                return result

            self.prompt()
            self.prompt(error_message)

    def text_input(self, prompt: str, default: str | None = None) -> str:
        """Prompt the user for text input.

        If no default is specified, the input will be returned as entered.

        The default will also be returned if input is disabled. If input is disabled,
        and there is no default, InputDisabled will be raised.

        :param prompt: The prompt to display to the user.
        :param default: (optional) The response to return if the user provides no input.
        :returns: The content entered by the user.
        """
        try:
            user_input = self(prompt)
            if default is not None and user_input == "":
                return default
        except InputDisabled:
            if default is not None:
                return default
            raise

        return user_input

    def __call__(self, prompt: str, *, markup: bool = False):
        """Present input() interface; prompt should be bold if markup is included."""
        if not self.enabled:
            raise InputDisabled()

        # make the prompt *bold* if it doesn't already contain markup
        escaped_prompt = f"[bold]{escape(prompt)}[/bold]" if not markup else prompt

        try:
            input_value = self.input(escaped_prompt, markup=True)
        except EOFError:
            raise KeyboardInterrupt

        self.print.to_log(prompt)
        self.print.to_log(f"User input: {input_value}")

        return input_value


def select_option(options, input, prompt="> ", error="Invalid selection", default=None):
    """Prompt the user for a choice from a list of options.

    The options are provided as a dictionary; the values are the human-readable options,
    and the keys are the values that will be returned as the selection. The human-
    readable options will be sorted before display to the user.

    This method does *not* print a question or any leading text; it only prints the list
    of options, and prompts the user for their choice. If the user chooses an invalid
    selection (either provides non-integer input, or an invalid integer), it prints an
    error message and prompts the user again.

    :param options: A dictionary, or list of tuples, of options to present to the user.
    :param input: The function to use to retrieve the user's input. This exists so that
        the user's input can be easily mocked during testing.
    :param prompt: The prompt to display to the user.
    :param error: The error message to display when the user provides invalid input.
    :param default: The default option for empty user input. The options for the user
        start numbering at 1; so, to default to the first item, this should be "1".
    :returns: The key corresponding to the user's chosen option.
    """
    if isinstance(options, dict):
        ordered = list(sorted(options.items(), key=operator.itemgetter(1)))
    else:
        ordered = options

    if input.enabled:
        for i, (key, value) in enumerate(ordered, start=1):
            input.prompt(f"  {i}) {value}")

        input.prompt()

    choices = [str(index) for index in range(1, len(ordered) + 1)]
    index = input.selection_input(
        prompt=prompt, choices=choices, error_message=error, default=default
    )
    return ordered[int(index) - 1][0]



from ...utils import assert_url_resolvable


def test_managed_install(mock_tools, rcedit):
    """All rcedit installs are managed."""
    assert rcedit.managed_install


def test_rcedit_url(mock_tools, rcedit):
    """The URL for RCEdit is correct."""
    assert (
        rcedit.download_url
        == "https://github.com/electron/rcedit/releases/download/v2.0.0/rcedit-x64.exe"
    )
    assert_url_resolvable(rcedit.download_url)



import os
from unittest.mock import MagicMock

import pytest

from briefcase.exceptions import (
    BriefcaseCommandError,
    MissingToolError,
    NetworkFailure,
    UnsupportedHostError,
)
from briefcase.integrations.wix import WiX

from ...utils import assert_url_resolvable, create_zip_file
from .conftest import WIX_DOWNLOAD_URL


def test_short_circuit(mock_tools):
    """Tool is not created if already cached."""
    mock_tools.wix = "tool"

    tool = WiX.verify(mock_tools)

    assert tool == "tool"
    assert tool == mock_tools.wix


@pytest.mark.parametrize("host_os", ["Darwin", "Linux", "wonky"])
def test_unsupported_os(mock_tools, host_os):
    """When host OS is not supported, an error is raised."""
    mock_tools.host_os = host_os

    with pytest.raises(
        UnsupportedHostError,
        match=f"{WiX.name} is not supported on {host_os}",
    ):
        WiX.verify(mock_tools)


def test_valid_wix_envvar(mock_tools, tmp_path):
    """If the WiX envvar points to a valid WiX install, the validator succeeds."""
    # Mock the environment for a WiX install
    wix_path = tmp_path / "wix"
    mock_tools.os.environ.get.return_value = os.fsdecode(wix_path)

    # Mock the interesting parts of a WiX install
    (wix_path / "bin").mkdir(parents=True)
    (wix_path / "bin/heat.exe").touch()
    (wix_path / "bin/light.exe").touch()
    (wix_path / "bin/candle.exe").touch()

    # Verify the install
    wix = WiX.verify(mock_tools)

    # The environment was queried.
    mock_tools.os.environ.get.assert_called_with("WIX")

    # The returned paths are as expected (and are the full paths)
    assert wix.heat_exe == tmp_path / "wix/bin/heat.exe"
    assert wix.light_exe == tmp_path / "wix/bin/light.exe"
    assert wix.candle_exe == tmp_path / "wix/bin/candle.exe"


def test_invalid_wix_envvar(mock_tools, tmp_path):
    """If the WiX envvar points to an invalid WiX install, the validator fails."""
    # Mock the environment for a WiX install
    wix_path = tmp_path / "wix"
    mock_tools.os.environ.get.return_value = os.fsdecode(wix_path)

    # Don't create the actual install, and then attempt to validate
    with pytest.raises(BriefcaseCommandError, match="does not point to an install"):
        WiX.verify(mock_tools)


def test_existing_wix_install(mock_tools, tmp_path):
    """If there's an existing managed WiX install, the validator succeeds."""
    # Mock the environment as if there is not WiX variable
    mock_tools.os.environ.get.return_value = None

    # Create a mock of a previously installed WiX version.
    wix_path = tmp_path / "tools/wix"
    wix_path.mkdir(parents=True)
    (wix_path / "heat.exe").touch()
    (wix_path / "light.exe").touch()
    (wix_path / "candle.exe").touch()

    wix = WiX.verify(mock_tools)

    # The environment was queried.
    mock_tools.os.environ.get.assert_called_with("WIX")

    # No download was attempted
    assert mock_tools.file.download.call_count == 0

    # The returned paths are as expected
    assert wix.heat_exe == tmp_path / "tools/wix/heat.exe"
    assert wix.light_exe == tmp_path / "tools/wix/light.exe"
    assert wix.candle_exe == tmp_path / "tools/wix/candle.exe"


def test_download_wix(mock_tools, tmp_path):
    """If there's no existing managed WiX install, it is downloaded and unpacked."""
    # Mock the environment as if there is not WiX variable
    mock_tools.os.environ.get.return_value = None

    # Mock the download
    wix_path = tmp_path / "tools/wix"

    wix_zip_path = create_zip_file(tmp_path / "tools/wix.zip", content=[("wix", "wix")])

    mock_tools.file.download = MagicMock(return_value=wix_zip_path)

    # Verify the install. This will trigger a download
    wix = WiX.verify(mock_tools)

    # The environment was queried.
    mock_tools.os.environ.get.assert_called_with("WIX")

    # A download was initiated
    mock_tools.file.download.assert_called_with(
        url=WIX_DOWNLOAD_URL,
        download_path=tmp_path / "tools",
        role="WiX",
    )

    # The download was unpacked.
    mock_tools.shutil.unpack_archive.assert_called_with(
        filename=os.fsdecode(wix_zip_path), extract_dir=os.fsdecode(wix_path)
    )

    # The zip file was removed
    assert not wix_zip_path.exists()

    # The returned paths are as expected
    assert wix.heat_exe == tmp_path / "tools/wix/heat.exe"
    assert wix.light_exe == tmp_path / "tools/wix/light.exe"
    assert wix.candle_exe == tmp_path / "tools/wix/candle.exe"

    # The WiX URL is resolvable
    assert_url_resolvable(wix.download_url)


def test_dont_install(mock_tools, tmp_path):
    """If there's no existing managed WiX install, an install is not requested, verify
    fails."""
    # Mock the environment as if there is not WiX variable
    mock_tools.os.environ.get.return_value = None

    # Verify, but don't install. This will fail.
    with pytest.raises(MissingToolError):
        WiX.verify(mock_tools, install=False)

    # The environment was queried.
    mock_tools.os.environ.get.assert_called_with("WIX")

    # No download was initiated
    mock_tools.file.download.assert_not_called()


def test_download_fail(mock_tools, tmp_path):
    """If the download doesn't complete, the validator fails."""
    # Mock the environment as if there is not WiX variable
    mock_tools.os.environ.get.return_value = None

    # Mock the download failure
    mock_tools.file.download = MagicMock(side_effect=NetworkFailure("mock"))

    # Verify the install. This will trigger a download
    with pytest.raises(NetworkFailure, match="Unable to mock"):
        WiX.verify(mock_tools)

    # The environment was queried.
    mock_tools.os.environ.get.assert_called_with("WIX")

    # A download was initiated
    mock_tools.file.download.assert_called_with(
        url=WIX_DOWNLOAD_URL,
        download_path=tmp_path / "tools",
        role="WiX",
    )

    # ... but the unpack didn't happen
    assert mock_tools.shutil.unpack_archive.call_count == 0


def test_unpack_fail(mock_tools, tmp_path):
    """If the download archive is corrupted, the validator fails."""
    # Mock the environment as if there is not WiX variable
    mock_tools.os.environ.get.return_value = None

    # Mock the download
    wix_path = tmp_path / "tools/wix"

    wix_zip_path = create_zip_file(tmp_path / "tools/wix.zip", content=[("wix", "wix")])

    mock_tools.file.download = MagicMock(return_value=wix_zip_path)

    # Mock an unpack failure
    mock_tools.shutil.unpack_archive.side_effect = EOFError

    # Verify the install. This will trigger a download,
    # but the unpack will fail
    with pytest.raises(BriefcaseCommandError, match="interrupted or corrupted"):
        WiX.verify(mock_tools)

    # The environment was queried.
    mock_tools.os.environ.get.assert_called_with("WIX")

    # A download was initiated
    mock_tools.file.download.assert_called_with(
        url=WIX_DOWNLOAD_URL,
        download_path=tmp_path / "tools",
        role="WiX",
    )

    # The download was unpacked.
    mock_tools.shutil.unpack_archive.assert_called_with(
        filename=os.fsdecode(wix_zip_path), extract_dir=os.fsdecode(wix_path)
    )

    # The zip file was not removed
    assert wix_zip_path.exists()



import subprocess

import pytest

from briefcase.exceptions import BriefcaseCommandError


def test_verify_license_passes_quickly_if_license_present(mock_tools, android_sdk):
    """Validate that verify_license() successfully does nothing in its happy path.

    If `android-sdk-license` exists in the right place, we expect verify_license() to
    run no subprocesses.
    """
    license_path = android_sdk.root_path / "licenses/android-sdk-license"
    license_path.parent.mkdir(parents=True)
    license_path.touch()

    android_sdk.verify_license()
    mock_tools.subprocess.run.assert_not_called()


def test_verify_license_prompts_for_licenses_and_exits_if_you_agree(
    mock_tools,
    android_sdk,
):
    """Validate that if verify_license() succeeds if you agree to the Android SDK
    license."""

    def accept_license(*args, **kwargs):
        license_dir = android_sdk.root_path / "licenses"
        license_dir.mkdir(parents=True)
        (license_dir / "android-sdk-license").touch()

    mock_tools.subprocess.run.side_effect = accept_license
    android_sdk.verify_license()
    mock_tools.subprocess.run.assert_called_once_with(
        [android_sdk.sdkmanager_path, "--licenses"],
        env=android_sdk.env,
        check=True,
        stream_output=False,
    )


def test_verify_license_handles_sdkmanager_crash(mock_tools, android_sdk):
    """Validate that if verify_license() raises a briefcase exception if it fails to
    launch `sdkmanager`."""
    mock_tools.subprocess.run.side_effect = subprocess.CalledProcessError(1, "")
    with pytest.raises(BriefcaseCommandError):
        android_sdk.verify_license()

    mock_tools.subprocess.run.assert_called_once_with(
        [android_sdk.sdkmanager_path, "--licenses"],
        env=android_sdk.env,
        check=True,
        stream_output=False,
    )


def test_verify_license_insists_on_agreement(mock_tools, android_sdk):
    """Validate that if the user quits `sdkmanager --licenses` without agreeing to the
    license, verify_license() raises an exception."""
    # Simulate user non-acceptance of the license by allowing the mock
    # subprocess.run() to take no action.
    with pytest.raises(BriefcaseCommandError):
        android_sdk.verify_license()

    mock_tools.subprocess.run.assert_called_once_with(
        [android_sdk.sdkmanager_path, "--licenses"],
        env=android_sdk.env,
        check=True,
        stream_output=False,
    )



from __future__ import annotations

import hashlib
import os
import platform
import shutil
import subprocess
import sys
from datetime import date, datetime
from pathlib import Path

import briefcase
from briefcase.config import AppConfig
from briefcase.exceptions import (
    BriefcaseCommandError,
    InvalidStubBinary,
    InvalidSupportPackage,
    MissingAppSources,
    MissingNetworkResourceError,
    MissingStubBinary,
    MissingSupportPackage,
    RequirementsInstallError,
    UnsupportedPlatform,
)
from briefcase.integrations.git import Git
from briefcase.integrations.subprocess import NativeAppContext

from .base import BaseCommand, full_options


def cookiecutter_cache_path(template):
    """Determine the cookiecutter template cache directory given a template URL.

    This will return a valid path, regardless of whether `template`

    :param template: The template to use. This can be a filesystem path or
        a URL.
    :returns: The path that cookiecutter would use for the given template name.
    """
    template = template.rstrip("/")
    tail = template.split("/")[-1]
    cache_name = tail.rsplit(".git")[0]
    return Path.home() / ".cookiecutters" / cache_name


def write_dist_info(app: AppConfig, dist_info_path: Path):
    """Install the dist-info folder for the application.

    :param app: The config object for the app
    :param dist_info_path: The path into which the dist-info folder should be written.
    """
    # Create dist-info folder, and write a minimal metadata collection.
    dist_info_path.mkdir(exist_ok=True)
    with (dist_info_path / "INSTALLER").open("w", encoding="utf-8") as f:
        f.write("briefcase\n")
    with (dist_info_path / "WHEEL").open("w", encoding="utf-8") as f:
        f.write("Wheel-Version: 1.0\n")
        f.write("Root-Is-Purelib: true\n")
        f.write(f"Generator: briefcase ({briefcase.__version__})\n")
        f.write("Tag: py3-none-any\n")
    with (dist_info_path / "METADATA").open("w", encoding="utf-8") as f:
        f.write("Metadata-Version: 2.1\n")
        f.write(f"Briefcase-Version: {briefcase.__version__}\n")
        f.write(f"Name: {app.app_name}\n")
        f.write(f"Formal-Name: {app.formal_name}\n")
        f.write(f"App-ID: {app.bundle_identifier}\n")
        f.write(f"Version: {app.version}\n")
        if app.url:
            f.write(f"Home-page: {app.url}\n")
            f.write(f"Download-URL: {app.url}\n")
        else:
            f.write("Download-URL: \n")
        if app.author:
            f.write(f"Author: {app.author}\n")
        if app.author_email:
            f.write(f"Author-email: {app.author_email}\n")
        f.write(f"Summary: {app.description}\n")
    with (dist_info_path / "top_level.txt").open("w", encoding="utf-8") as f:
        f.write(f"{app.module_name}\n")


class CreateCommand(BaseCommand):
    command = "create"
    description = "Create a new app for a target platform."

    # app properties that won't be exposed to the context
    hidden_app_properties = {"permission"}

    @property
    def app_template_url(self) -> str:
        """The URL for a cookiecutter repository to use when creating apps."""
        return f"https://github.com/beeware/briefcase-{self.platform}-{self.output_format}-template.git"

    def support_package_filename(self, support_revision: str) -> str:
        """The query arguments to use in a support package query request."""
        return f"Python-{self.python_version_tag}-{self.platform}-support.b{support_revision}.tar.gz"

    def support_package_url(self, support_revision: str) -> str:
        """The URL of the support package to use for apps of this type."""
        return (
            "https://briefcase-support.s3.amazonaws.com/python/"
            f"{self.python_version_tag}/"
            f"{self.platform}/"
            f"{self.support_package_filename(support_revision)}"
        )

    def stub_binary_filename(self, support_revision: str, is_console_app: bool) -> str:
        """The filename for the stub binary."""
        stub_type = "Console" if is_console_app else "GUI"
        return f"{stub_type}-Stub-{self.python_version_tag}-b{support_revision}.zip"

    def stub_binary_url(self, support_revision: str, is_console_app: bool) -> str:
        """The URL of the stub binary to use for apps of this type."""
        return (
            "https://briefcase-support.s3.amazonaws.com/python/"
            f"{self.python_version_tag}/"
            f"{self.platform}/"
            f"{self.stub_binary_filename(support_revision, is_console_app)}"
        )

    def icon_targets(self, app: AppConfig):
        """Obtain the dictionary of icon targets that the template requires.

        :param app: The config object for the app
        :return: A dictionary of icons that the template supports. The keys of the
            dictionary are the size of the icons.
        """
        # If the template specifies no icons, return an empty dictionary.
        # If the template specifies a single icon without a size specification,
        #   return a dictionary with a single ``None`` key.
        # Otherwise, return the full size-keyed dictionary.
        try:
            icon_targets = self.path_index(app, "icon")
            # Convert string-specified icons into an "unknown size" icon form
            if isinstance(icon_targets, str):
                icon_targets = {None: icon_targets}
        except KeyError:
            icon_targets = {}

        return icon_targets

    def document_type_icon_targets(self, app: AppConfig):
        """Obtain the dictionary of document type icon targets that the template
        requires.

        :param app: The config object for the app
        :return: A dictionary of document types, with the values being dictionaries
            describing the icon sizes that the template supports. The inner dictionary
            describes the path fragments (relative to the bundle path) for the images
            that are required; the keys are the size of the splash images.
        """
        # If the template specifies no document types, return an empty dictionary.
        # Then, for each document type; If the template specifies a single icon
        #   without a size specification, return a dictionary with a single
        #   ``None`` key. Otherwise, return the full size-keyed dictionary.
        try:
            icon_targets = self.path_index(app, "document_type_icon")
            return {
                extension: {None: targets} if isinstance(targets, str) else targets
                for extension, targets in icon_targets.items()
            }

        except KeyError:
            return {}

    def _x_permissions(self, app: AppConfig):
        """Extract the known cross-platform permission definitions from the app's
        permissions definitions.

        After calling this method, the ``permissions`` declaration for the app will
        only contain keys that are *not* cross-platform keys.

        :param app: The config object for the app
        :returns: A dictionary of known cross-platform permission definitions.
        """
        return {
            key: app.permission.pop(key, None)
            for key in [
                "camera",
                "microphone",
                "coarse_location",
                "fine_location",
                "background_location",
                "photo_library",
            ]
        }

    def permissions_context(self, app: AppConfig, x_permissions: dict[str, str]):
        """Additional template context for permissions.

        :param app: The config object for the app
        :param x_permissions: The dictionary of known cross-platform permission
            definitions.
        :returns: The template context describing permissions for the app.
        """
        return {}

    def output_format_template_context(self, app: AppConfig):
        """Additional template context required by the output format.

        :param app: The config object for the app
        """
        return {}

    def generate_app_template(self, app: AppConfig):
        """Create an application bundle.

        :param app: The config object for the app
        """
        # If the app config doesn't explicitly define a template,
        # use a default template.

        # Construct a template context from the app configuration.
        extra_context = {
            key: value
            for key, value in app.__dict__.items()
            if key not in self.hidden_app_properties
        }

        # Remove the context items that describe the template
        extra_context.pop("template")
        extra_context.pop("template_branch")

        # Augment with some extra fields.
        extra_context.update(
            {
                # Ensure the output format is in the case we expect
                "format": self.output_format.lower(),
                # Properties of the generating environment
                # The full Python version string, including minor and dev/a/b/c suffixes (e.g., 3.11.0rc2)
                "python_version": platform.python_version(),
                # The host architecture
                "host_arch": self.tools.host_arch,
                # Transformations of explicit properties into useful forms
                "class_name": app.class_name,
                "module_name": app.module_name,
                "package_name": app.package_name,
                "bundle_identifier": app.bundle_identifier,
                # Properties that are a function of the execution
                "year": date.today().strftime("%Y"),
                "month": date.today().strftime("%B"),
            }
        )

        # Add in any extra template context to support permissions
        extra_context.update(self.permissions_context(app, self._x_permissions(app)))

        # Add in any extra template context required by the output format.
        extra_context.update(self.output_format_template_context(app))

        # Create the platform directory (if it doesn't already exist)
        output_path = self.bundle_path(app).parent
        output_path.mkdir(parents=True, exist_ok=True)

        self.generate_template(
            template=app.template if app.template else self.app_template_url,
            branch=app.template_branch,
            output_path=output_path,
            extra_context=extra_context,
        )

    def _unpack_support_package(self, support_file_path, support_path):
        """Unpack a support package into a specific location.

        :param support_file_path: The path to the support file to be unpacked.
        :param support_path: The path where support files should be unpacked.
        """
        try:
            with self.input.wait_bar("Unpacking support package..."):
                support_path.mkdir(parents=True, exist_ok=True)
                self.tools.file.unpack_archive(
                    support_file_path,
                    extract_dir=support_path,
                )
        except (shutil.ReadError, EOFError) as e:
            raise InvalidSupportPackage(support_file_path) from e

    def _cleanup_app_support_package(self, support_path):
        """The internal implementation of the app support cleanup method.

        Guaranteed to only be invoked if the backend uses a support package, and the
        support path exists.

        :param support_path: The support path to clean up.
        """
        with self.input.wait_bar("Removing existing support package..."):
            self.tools.shutil.rmtree(support_path)

    def cleanup_app_support_package(self, app: AppConfig):
        """Clean up an existing application support package.

        :param app: The config object for the app
        """
        try:
            support_path = self.support_path(app)
        except KeyError:
            # No cleanup required.
            pass
        else:
            if support_path.exists():
                self._cleanup_app_support_package(support_path)

    def install_app_support_package(self, app: AppConfig):
        """Install the application support package.

        :param app: The config object for the app
        """
        try:
            support_path = self.support_path(app)
        except KeyError:
            self.logger.info("No support package required.")
        else:
            support_file_path = self._download_support_package(app)
            self._unpack_support_package(support_file_path, support_path)

    def _download_support_package(self, app: AppConfig):
        try:
            # Work out if the app defines a custom override for
            # the support package URL.
            try:
                support_package_url = app.support_package
                custom_support_package = True
                self.logger.info(f"Using custom support package {support_package_url}")
                try:
                    # If the app has a custom support package *and* a support revision,
                    # that's an error.
                    app.support_revision
                    self.logger.warning(
                        "App specifies both a support package and a support revision; "
                        "support revision will be ignored."
                    )
                except AttributeError:
                    pass
            except AttributeError:
                # If the app specifies a support revision, use it;
                # otherwise, use the support revision named by the template
                try:
                    support_revision = app.support_revision
                except AttributeError:
                    # No support revision specified; use the template-specified version
                    try:
                        support_revision = self.support_revision(app)
                    except KeyError:
                        # No template-specified support revision
                        raise MissingSupportPackage(
                            python_version_tag=self.python_version_tag,
                            platform=self.platform,
                            host_arch=self.tools.host_arch,
                            is_32bit=self.tools.is_32bit_python,
                        )

                support_package_url = self.support_package_url(support_revision)
                custom_support_package = False
                self.logger.info(f"Using support package {support_package_url}")

            if support_package_url.startswith(("https://", "http://")):
                if custom_support_package:
                    # If the support package is custom, cache it using a hash of
                    # the download URL. This is needed to differentiate to support
                    # packages with the same filename, served at different URLs.
                    # (or a custom package that collides with an official package name)
                    download_path = (
                        self.data_path
                        / "support"
                        / hashlib.sha256(
                            support_package_url.encode("utf-8")
                        ).hexdigest()
                    )
                else:
                    download_path = self.data_path / "support"

                # Download the support file, caching the result
                # in the user's briefcase support cache directory.
                return self.tools.file.download(
                    url=support_package_url,
                    download_path=download_path,
                    role="support package",
                )
            else:
                return Path(support_package_url)
        except MissingNetworkResourceError as e:
            # If there is a custom support package, report the missing resource as-is.
            if custom_support_package:
                raise
            else:
                raise MissingSupportPackage(
                    python_version_tag=self.python_version_tag,
                    platform=self.platform,
                    host_arch=self.tools.host_arch,
                    is_32bit=self.tools.is_32bit_python,
                ) from e

    def cleanup_stub_binary(self, app: AppConfig):
        """Clean up an existing application support package.

        :param app: The config object for the app
        """
        with self.input.wait_bar("Removing existing stub binary..."):
            self.binary_executable_path(app).unlink(missing_ok=True)
            self.unbuilt_executable_path(app).unlink(missing_ok=True)

    def install_stub_binary(self, app: AppConfig):
        """Install the application stub binary into the "unbuilt" location.

        :param app: The config object for the app
        """
        unbuilt_executable_path = self.unbuilt_executable_path(app)
        stub_binary_path = self._download_stub_binary(app)

        with self.input.wait_bar("Installing stub binary..."):
            # Ensure the folder for the stub binary exists
            unbuilt_executable_path.parent.mkdir(exist_ok=True, parents=True)

            # Install the stub binary into the unbuilt location.
            # Allow for both raw and compressed artefacts.
            try:
                if self.tools.file.is_archive(stub_binary_path):
                    self.tools.file.unpack_archive(
                        stub_binary_path,
                        extract_dir=unbuilt_executable_path.parent,
                    )
                elif stub_binary_path.is_file():
                    self.tools.shutil.copyfile(
                        stub_binary_path, unbuilt_executable_path
                    )
                else:
                    raise InvalidStubBinary(stub_binary_path)
            except (shutil.ReadError, EOFError, OSError) as e:
                raise InvalidStubBinary(stub_binary_path) from e
            else:
                # Ensure the binary is executable
                self.tools.os.chmod(unbuilt_executable_path, 0o755)

    def _download_stub_binary(self, app: AppConfig) -> Path:
        try:
            # Work out if the app defines a custom override for
            # the support package URL.
            try:
                stub_binary_url = app.stub_binary
                custom_stub_binary = True
                self.logger.info(f"Using custom stub binary {stub_binary_url}")
                try:
                    # If the app has a custom stub binary *and* a support revision,
                    # that's an error.
                    app.stub_binary_revision
                    self.logger.warning(
                        "App specifies both a stub binary and a stub binary revision; "
                        "stub binary revision will be ignored."
                    )
                except AttributeError:
                    pass
            except AttributeError:
                # If the app specifies a support revision, use it; otherwise, use the
                # support revision named by the template. This value *must* exist, as
                # stub binary handling won't be triggered at all unless it is present.
                try:
                    stub_binary_revision = app.stub_binary_revision
                except AttributeError:
                    stub_binary_revision = self.stub_binary_revision(app)

                stub_binary_url = self.stub_binary_url(
                    stub_binary_revision, app.console_app
                )
                custom_stub_binary = False
                self.logger.info(f"Using stub binary {stub_binary_url}")

            if stub_binary_url.startswith(("https://", "http://")):
                if custom_stub_binary:
                    # If the support package is custom, cache it using a hash of
                    # the download URL. This is needed to differentiate to support
                    # packages with the same filename, served at different URLs.
                    # (or a custom package that collides with an official package name)
                    download_path = (
                        self.data_path
                        / "stub"
                        / hashlib.sha256(stub_binary_url.encode("utf-8")).hexdigest()
                    )
                else:
                    download_path = self.data_path / "stub"

                # Download the stub binary, caching the result
                # in the user's briefcase stub cache directory.
                return self.tools.file.download(
                    url=stub_binary_url,
                    download_path=download_path,
                    role="stub binary",
                )
            else:
                return Path(stub_binary_url)
        except MissingNetworkResourceError as e:
            # If there is a custom support package, report the missing resource as-is.
            if custom_stub_binary:
                raise
            else:
                raise MissingStubBinary(
                    python_version_tag=self.python_version_tag,
                    platform=self.platform,
                    host_arch=self.tools.host_arch,
                    is_32bit=self.tools.is_32bit_python,
                ) from e

    def _write_requirements_file(
        self,
        app: AppConfig,
        requires: list[str],
        requirements_path: Path,
    ):
        """Configure application requirements by writing a requirements.txt file.

        :param app: The app configuration
        :param requires: The full list of requirements
        :param requirements_path: The full path to a requirements.txt file that will be
            written.
        """

        with self.input.wait_bar("Writing requirements file..."):
            with requirements_path.open("w", encoding="utf-8") as f:
                if requires:
                    # Add timestamp so build systems (such as Gradle) detect a change
                    # in the file and perform a re-installation of all requirements.
                    f.write(f"# Generated {datetime.now()}\n")
                    for requirement in requires:
                        # If the requirement is a local path, convert it to
                        # absolute, because Flatpak moves the requirements file
                        # to a different place before using it.
                        if _is_local_requirement(requirement):
                            # We use os.path.abspath() rather than Path.resolve()
                            # because we *don't* want Path's symlink resolving behavior.
                            requirement = os.path.abspath(self.base_path / requirement)
                        f.write(f"{requirement}\n")

    def _pip_requires(self, app: AppConfig, requires: list[str]):
        """Convert the list of requirements to be passed to pip into its final form.

        :param app: The app configuration
        :param requires: The user-specified list of app requirements
        :returns: The final list of requirement arguments to pass to pip
        """
        return requires

    def _extra_pip_args(self, app: AppConfig):
        """Any additional arguments that must be passed to pip when installing packages.

        :param app: The app configuration
        :returns: A list of additional arguments
        """
        return []

    def _pip_install(
        self,
        app: AppConfig,
        app_packages_path: Path,
        pip_args: list[str],
        install_hint: str = "",
        **pip_kwargs,
    ):
        """Invoke pip to install a set of requirements.

        :param app: The app configuration
        :param app_packages_path: The full path of the app_packages folder into which
            requirements should be installed.
        :param pip_args: The list of arguments (including the list of requirements to
            install) to pass to pip. This is in addition to the default arguments that
            disable pip version checks, forces upgrades, and installs into the nominated
            ``app_packages`` path.
        :param install_hint: Additional hint information to provide in the exception
            message if the pip install call fails.
        :param pip_kwargs: Any additional keyword arguments to pass to ``subprocess.run``
            when invoking pip.
        """
        try:
            self.tools[app].app_context.run(
                [
                    sys.executable,
                    "-u",
                    "-X",
                    "utf8",
                    "-m",
                    "pip",
                    "install",
                    "--disable-pip-version-check",
                    "--no-python-version-warning",
                    "--upgrade",
                    "--no-user",
                    f"--target={app_packages_path}",
                ]
                + (["-vv"] if self.logger.is_deep_debug else [])
                + self._extra_pip_args(app)
                + pip_args,
                check=True,
                encoding="UTF-8",
                **pip_kwargs,
            )
        except subprocess.CalledProcessError as e:
            raise RequirementsInstallError(install_hint=install_hint) from e

    def _install_app_requirements(
        self,
        app: AppConfig,
        requires: list[str],
        app_packages_path: Path,
        progress_message: str = "Installing app requirements...",
        pip_kwargs: dict[str, str] | None = None,
    ):
        """Install requirements for the app with pip.

        :param app: The app configuration
        :param requires: The list of requirements to install
        :param app_packages_path: The full path of the app_packages folder into which
            requirements should be installed.
        :param progress_message: The waitbar progress message to display to the user.
        :param pip_kwargs: Any additional keyword arguments to pass to the subprocess
            when invoking pip.
        """
        # Clear existing dependency directory
        if app_packages_path.is_dir():
            self.tools.shutil.rmtree(app_packages_path)
            self.tools.os.mkdir(app_packages_path)

        # Install requirements
        if requires:
            with self.input.wait_bar(progress_message):
                self._pip_install(
                    app,
                    app_packages_path=app_packages_path,
                    pip_args=self._pip_requires(app, requires),
                    **(pip_kwargs if pip_kwargs else {}),
                )
        else:
            self.logger.info("No application requirements.")

    def install_app_requirements(self, app: AppConfig, test_mode: bool):
        """Handle requirements for the app.

        This will result in either (in preferential order):
         * a requirements.txt file being written at a location specified by
           ``app_requirements_path`` in the template path index
         * requirements being installed with pip into the location specified
           by the ``app_packages_path`` in the template path index.

        If ``test_mode`` is True, the test requirements will also be installed.

        If the path index doesn't specify either of the path index entries,
        an error is raised.

        :param app: The config object for the app
        :param test_mode: Should the test requirements be installed?
        """
        requires = app.requires.copy() if app.requires else []
        if test_mode and app.test_requires:
            requires.extend(app.test_requires)

        try:
            requirements_path = self.app_requirements_path(app)
            self._write_requirements_file(app, requires, requirements_path)
        except KeyError:
            try:
                app_packages_path = self.app_packages_path(app)
                self._install_app_requirements(app, requires, app_packages_path)
            except KeyError as e:
                raise BriefcaseCommandError(
                    "Application path index file does not define "
                    "`app_requirements_path` or `app_packages_path`"
                ) from e

    def install_app_code(self, app: AppConfig, test_mode: bool):
        """Install the application code into the bundle.

        :param app: The config object for the app
        :param test_mode: Should the application test code also be installed?
        """
        # Remove existing app folder if it exists
        app_path = self.app_path(app)
        if app_path.exists():
            self.tools.shutil.rmtree(app_path)
        self.tools.os.mkdir(app_path)

        sources = app.sources.copy() if app.sources else []
        if test_mode and app.test_sources:
            sources.extend(app.test_sources)

        # Install app code.
        if sources:
            for src in sources:
                with self.input.wait_bar(f"Installing {src}..."):
                    original = self.base_path / src
                    target = app_path / original.name

                    # Install the new copy of the app code.
                    if not original.exists():
                        raise MissingAppSources(src)
                    elif original.is_dir():
                        self.tools.shutil.copytree(original, target)
                    else:
                        self.tools.shutil.copy(original, target)
        else:
            self.logger.info(f"No sources defined for {app.app_name}.")

        # Write the dist-info folder for the application.
        write_dist_info(
            app=app,
            dist_info_path=self.app_path(app)
            / f"{app.module_name}-{app.version}.dist-info",
        )

    def install_image(self, role, variant, size, source, target):
        """Install an icon/image of the requested size at a target location, using the
        source images defined by the app config.

        :param role: A string describing the role the of the image.
        :param variant: The image variant. A variant of ``None`` means the image
            has no variants
        :param size: The requested size for the image. A size of
            ``None`` means the largest available size should be used.
        :param source: The image source; or a dictionary of sources for each
            variant. The sources will *not* include any extension or size
            modifier; these will be added based on the requested target and
            variant.
        :param target: The full path where the image should be installed.
        """
        if source is not None:
            if size is None:
                if variant is None:
                    source_filename = f"{source}{target.suffix}"
                    full_role = role
                else:
                    try:
                        full_role = f"{variant} {role}"
                        source_filename = f"{source[variant]}{target.suffix}"
                    except TypeError:
                        source_filename = f"{source}-{variant}{target.suffix}"
                    except KeyError:
                        self.logger.info(
                            f"Unknown variant {variant!r} for {role}; using default"
                        )
                        return
            else:
                if variant is None:
                    # An annoying edge case is the case of an unsized variant.
                    # In that case, `size` is actually the variant, and the
                    # source is a dictionary keyed by variant. Try that
                    # lookup; if it fails, we have a sized image with no
                    # variant.
                    try:
                        source_filename = f"{source[size]}{target.suffix}"
                        full_role = f"{size} {role}"
                    except TypeError:
                        # The lookup on the source failed; that means we
                        # have a sized image without variants.
                        source_filename = f"{source}-{size}{target.suffix}"
                        full_role = f"{size}px {role}"
                else:
                    try:
                        full_role = f"{size}px {variant} {role}"
                        source_filename = f"{source[variant]}-{size}{target.suffix}"
                    except TypeError:
                        source_filename = f"{source}-{variant}-{size}{target.suffix}"
                    except KeyError:
                        self.logger.info(
                            f"Unknown variant {variant!r} for {size}px {role}; using default"
                        )
                        return

            full_source = self.base_path / source_filename
            if full_source.exists():
                with self.input.wait_bar(
                    f"Installing {source_filename} as {full_role}..."
                ):
                    # Make sure the target directory exists
                    target.parent.mkdir(parents=True, exist_ok=True)
                    # Copy the source image to the target location
                    self.tools.shutil.copy(full_source, target)
            else:
                self.logger.info(
                    f"Unable to find {source_filename} for {full_role}; using default"
                )

    def install_app_resources(self, app: AppConfig):
        """Install the application resources (such as icons and splash screens) into the
        bundle.

        :param app: The config object for the app
        """
        for variant_or_size, targets in self.icon_targets(app).items():
            try:
                # Treat the targets as a dictionary of sizes;
                # if there's no `items`, then it's an icon without variants.
                for size, target in targets.items():
                    self.install_image(
                        "application icon",
                        source=app.icon,
                        variant=variant_or_size,
                        size=size,
                        target=self.bundle_path(app) / target,
                    )
            except AttributeError:
                # Either a single variant, or a single size.
                self.install_image(
                    "application icon",
                    source=app.icon,
                    variant=None,
                    size=variant_or_size,
                    target=self.bundle_path(app) / targets,
                )

        # Briefcase v0.3.18 - splash screens deprecated.
        if getattr(app, "splash", None):
            self.logger.warning()
            self.logger.warning(
                "Splash screens are now configured based on the icon. "
                "The splash configuration will be ignored."
            )

        for extension, doctype in self.document_type_icon_targets(app).items():
            self.logger.info()
            for size, target in doctype.items():
                self.install_image(
                    f"icon for .{extension} documents",
                    size=size,
                    source=app.document_types[extension]["icon"],
                    variant=None,
                    target=self.bundle_path(app) / target,
                )

    def cleanup_app_content(self, app: AppConfig):
        """Remove any content not needed by the final app bundle.

        :param app: The config object for the app
        """
        try:
            # Retrieve any cleanup paths defined by the app template
            paths_to_remove = self.cleanup_paths(app)
        except KeyError:
            paths_to_remove = []

        try:
            # Add any user-specified paths, expanded using the app as template context.
            paths_to_remove.extend([glob.format(app=app) for glob in app.cleanup_paths])
        except AttributeError:
            pass

        # Remove __pycache__ folders. These folders might contain stale PYC
        # artefacts, or encode file paths that reflect the original source
        # location. The stub binaries all disable PYC generation, to avoid
        # corrupting any app bundle signatures.
        paths_to_remove.append("**/__pycache__")

        with self.input.wait_bar("Removing unneeded app bundle content..."):
            for glob in paths_to_remove:
                # Expand each glob into a full list of files that actually exist
                # on the file system.
                for path in self.bundle_path(app).glob(glob):
                    relative_path = path.relative_to(self.bundle_path(app))
                    if path.is_dir():
                        self.logger.verbose(f"Removing directory {relative_path}")
                        self.tools.shutil.rmtree(path)
                    else:
                        self.logger.verbose(f"Removing {relative_path}")
                        path.unlink()

    def create_app(self, app: AppConfig, test_mode: bool = False, **options):
        """Create an application bundle.

        :param app: The config object for the app
        :param test_mode: Should the app be updated in test mode? (default: False)
        """
        if not app.supported:
            raise UnsupportedPlatform(self.platform)

        bundle_path = self.bundle_path(app)
        if bundle_path.exists():
            self.logger.info()
            confirm = self.input.boolean_input(
                f"Application {app.app_name!r} already exists; overwrite", default=False
            )
            if not confirm:
                self.logger.error(
                    f"Aborting creation of app {app.app_name!r}; existing application will not be overwritten."
                )
                return
            self.logger.info("Removing old application bundle...", prefix=app.app_name)
            self.tools.shutil.rmtree(bundle_path)

        self.logger.info("Generating application template...", prefix=app.app_name)
        self.generate_app_template(app=app)

        self.logger.info("Installing support package...", prefix=app.app_name)
        self.install_app_support_package(app=app)

        try:
            # If the platform uses a stub binary, the template will define a binary
            # revision. If this template configuration item doesn't exist, no stub
            # binary is required.
            self.stub_binary_revision(app)
        except KeyError:
            pass
        else:
            self.logger.info("Installing stub binary...", prefix=app.app_name)
            self.install_stub_binary(app=app)

        # Verify the app after the app template and support package
        # are in place since the app tools may be dependent on them.
        self.verify_app(app)

        self.logger.info("Installing application code...", prefix=app.app_name)
        self.install_app_code(app=app, test_mode=test_mode)

        self.logger.info("Installing requirements...", prefix=app.app_name)
        self.install_app_requirements(app=app, test_mode=test_mode)

        self.logger.info("Installing application resources...", prefix=app.app_name)
        self.install_app_resources(app=app)

        self.logger.info("Removing unneeded app content...", prefix=app.app_name)
        self.cleanup_app_content(app=app)

        self.logger.info(
            f"Created {bundle_path.relative_to(self.base_path)}",
            prefix=app.app_name,
        )

    def verify_tools(self):
        """Verify that the tools needed to run this command exist.

        Raises MissingToolException if a required system tool is missing.
        """
        super().verify_tools()
        Git.verify(tools=self.tools)

    def verify_app_tools(self, app: AppConfig):
        """Verify that tools needed to run the command for this app exist."""
        super().verify_app_tools(app)
        NativeAppContext.verify(tools=self.tools, app=app)

    def __call__(
        self,
        app: AppConfig | None = None,
        **options,
    ) -> dict | None:
        # Confirm host compatibility, that all required tools are available,
        # and that the app configuration is finalized.
        self.finalize(app)

        if app:
            state = self.create_app(app, **options)
        else:
            state = None
            for app_name, app in sorted(self.apps.items()):
                state = self.create_app(app, **full_options(state, options))

        return state


def _has_url(requirement):
    """Determine if the requirement is defined as a URL.

    Detects any of the URL schemes supported by pip
    (https://pip.pypa.io/en/stable/topics/vcs-support/).

    :param requirement: The requirement to check
    :returns: True if the requirement is a URL supported by pip.
    """
    return any(
        f"{scheme}:" in requirement
        for scheme in (
            ["http", "https", "file", "ftp"]
            + ["git+file", "git+https", "git+ssh", "git+http", "git+git", "git"]
            + ["hg+file", "hg+http", "hg+https", "hg+ssh", "hg+static-http"]
            + ["svn", "svn+svn", "svn+http", "svn+https", "svn+ssh"]
            + ["bzr+http", "bzr+https", "bzr+ssh", "bzr+sftp", "bzr+ftp", "bzr+lp"]
        )
    )


def _is_local_requirement(requirement):
    """Determine if the requirement is a local file path.

    :param requirement: The requirement to check
    :returns: True if the requirement is a local file path
    """
    # Windows allows both / and \ as a path separator in requirements.
    separators = [os.sep]
    if os.altsep:
        separators.append(os.altsep)

    return any(sep in requirement for sep in separators) and (not _has_url(requirement))



import tomli_w

try:
    import tomllib
except ModuleNotFoundError:
    import tomli as tomllib


def test_copy_pyproject_toml(convert_command, tmp_path):
    base_config_file = convert_command.base_path / "pyproject.toml"
    briefcase_config_file = tmp_path / "input" / "pyproject.toml"

    briefcase_config_file.parent.mkdir(parents=True)
    briefcase_config_file.write_text("placeholder", encoding="utf-8")
    convert_command.merge_or_copy_pyproject(briefcase_config_file)

    assert base_config_file.is_file()
    with open(base_config_file, encoding="utf-8") as file:
        assert file.read() == "placeholder"


def test_merge_pyproject(convert_command, tmp_path):
    base_config_file = convert_command.base_path / "pyproject.toml"
    briefcase_config_file = tmp_path / "input" / "pyproject.toml"

    base_config_content = {"placeholder1": "a"}
    briefcase_config_content = {"placeholder2": "b"}
    merged_dict = {**base_config_content, **briefcase_config_content}

    with open(base_config_file, "wb") as file:
        tomli_w.dump(base_config_content, file)

    briefcase_config_file.parent.mkdir(parents=True)
    with open(briefcase_config_file, "wb") as file:
        tomli_w.dump(briefcase_config_content, file)

    convert_command.merge_or_copy_pyproject(briefcase_config_file)

    merged_content = base_config_file.read_text(encoding="utf-8")
    assert merged_dict == tomllib.loads(merged_content)



import subprocess
from unittest.mock import MagicMock

import pytest

from briefcase.exceptions import BriefcaseCommandError


def test_kill(mock_tools, adb):
    """An emulator can be killed."""
    # Invoke kill
    adb.kill()

    # Validate call parameters.
    mock_tools.subprocess.check_output.assert_called_once_with(
        [
            mock_tools.android_sdk.adb_path,
            "-s",
            "exampleDevice",
            "emu",
            "kill",
        ],
        quiet=False,
    )


def test_kill_failure(adb):
    """If emu kill fails, the error is caught."""
    # Mock out the run command on an adb instance
    adb.run = MagicMock(
        side_effect=subprocess.CalledProcessError(returncode=1, cmd="adb emu kill")
    )

    with pytest.raises(BriefcaseCommandError):
        adb.kill()



import subprocess
from contextlib import suppress
from pathlib import Path

from briefcase.commands import BuildCommand, OpenCommand, PublishCommand, UpdateCommand
from briefcase.config import BaseConfig
from briefcase.exceptions import BriefcaseCommandError
from briefcase.integrations.rcedit import RCEdit
from briefcase.integrations.windows_sdk import WindowsSDK
from briefcase.platforms.windows import (
    WindowsCreateCommand,
    WindowsMixin,
    WindowsPackageCommand,
    WindowsRunCommand,
)


class WindowsAppMixin(WindowsMixin):
    output_format = "app"
    packaging_root = Path("src")

    def project_path(self, app):
        return self.bundle_path(app)


class WindowsAppCreateCommand(WindowsAppMixin, WindowsCreateCommand):
    description = "Create and populate a Windows app."


class WindowsAppUpdateCommand(WindowsAppCreateCommand, UpdateCommand):
    description = "Update an existing Windows app."


class WindowsAppOpenCommand(WindowsAppMixin, OpenCommand):
    description = "Open the folder containing an existing Windows app."


class WindowsAppBuildCommand(WindowsAppMixin, BuildCommand):
    description = "Build a Windows app."

    def verify_tools(self):
        super().verify_tools()
        RCEdit.verify(tools=self.tools)
        # The Windows SDK is only needed if it has previously been used to sign
        # the binary and MSI; therefore, ignore if it isn't available since the
        # stub app should not have any signatures to remove in that case.
        with suppress(BriefcaseCommandError):
            WindowsSDK.verify(tools=self.tools)

    def build_app(self, app: BaseConfig, **kwargs):
        """Build the application.

        :param app: The config object for the app
        """
        self.logger.info("Building App...", prefix=app.app_name)

        # Move the stub binary in to the final executable location
        unbuilt_binary_path = self.unbuilt_executable_path(app)
        if unbuilt_binary_path.exists():
            with self.input.wait_bar("Renaming stub binary..."):
                unbuilt_binary_path.rename(self.binary_executable_path(app))

        if hasattr(self.tools, "windows_sdk"):
            # If an app has been packaged and code signed previously, then the digital
            # signature on the app binary needs to be removed before re-building the app.
            # It is not safe to use RCEdit on signed binaries since it corrupts them.
            with self.input.wait_bar(
                "Removing any digital signatures from stub app..."
            ):
                try:
                    self.tools.subprocess.check_output(
                        [
                            self.tools.windows_sdk.signtool_exe,
                            "remove",
                            "-s",
                            self.binary_path(app).relative_to(self.bundle_path(app)),
                        ],
                        cwd=self.bundle_path(app),
                    )
                except subprocess.CalledProcessError as e:
                    # Ignore this error from signtool since it is logged if the file
                    # is not currently signed
                    if "error: 0x00000057" not in e.stdout:
                        raise BriefcaseCommandError(
                            f"""\
Failed to remove any existing digital signatures from the stub app.

Recreating the app layout may also help resolve this issue:

    $ briefcase create {self.platform} {self.output_format}

"""
                        ) from e

        with self.input.wait_bar("Setting stub app details..."):
            try:
                self.tools.subprocess.run(
                    [
                        self.tools.rcedit.rcedit_path,
                        self.binary_path(app).relative_to(self.bundle_path(app)),
                        "--set-version-string",
                        "CompanyName",
                        app.author,
                        # Although "FileDescription" sounds like it should be a... description,
                        # this is the label that appears as a grouping in the Task Manager
                        # when the application runs.
                        "--set-version-string",
                        "FileDescription",
                        app.formal_name,
                        "--set-version-string",
                        "FileVersion",
                        app.version,
                        "--set-version-string",
                        "InternalName",
                        app.module_name,
                        "--set-version-string",
                        "OriginalFilename",
                        self.binary_path(app).name,
                        "--set-version-string",
                        "ProductName",
                        app.formal_name,
                        "--set-version-string",
                        "ProductVersion",
                        app.version,
                        "--set-icon",
                        "icon.ico",
                    ],
                    check=True,
                    cwd=self.bundle_path(app),
                )
            except subprocess.CalledProcessError as e:
                raise BriefcaseCommandError(
                    f"Unable to update details on stub app for {app.app_name}."
                ) from e


class WindowsAppRunCommand(WindowsAppMixin, WindowsRunCommand):
    description = "Run a Windows app."


class WindowsAppPackageCommand(WindowsAppMixin, WindowsPackageCommand):
    description = "Package a Windows app."


class WindowsAppPublishCommand(WindowsAppMixin, PublishCommand):
    description = "Publish a Windows app."


# Declare the briefcase command bindings
create = WindowsAppCreateCommand
update = WindowsAppUpdateCommand
open = WindowsAppOpenCommand
build = WindowsAppBuildCommand
run = WindowsAppRunCommand
package = WindowsAppPackageCommand
publish = WindowsAppPublishCommand
